"""Various helpers for string quoting/unquoting.

Here is pure Python that should match C code in _cquoting.
"""

import re
from typing import Any, Dict, Mapping, Match, Optional
from urllib.parse import quote_plus, unquote_plus  # noqa

__all__ = (
    "quote_literal", "quote_copy", "quote_bytea_raw",
    "db_urlencode", "db_urldecode", "unescape",
    "unquote_literal",
)


#
# SQL quoting
#

def quote_literal(value: Any) -> str:
    r"""Quote a literal value for SQL.

    If string contains '\\', extended E'' quoting is used,
    otherwise standard quoting.  Input value of None results
    in string "null" without quotes.

    Python implementation.
    """

    if value is None:
        return "null"
    s = str(value).replace("'", "''")
    s2 = s.replace("\\", "\\\\")
    if len(s) != len(s2):
        return "E'" + s2 + "'"
    return "'" + s2 + "'"


def quote_copy(value: Any) -> str:
    """Quoting for copy command.  None is converted to \\N.

    Python implementation.
    """

    if value is None:
        return "\\N"
    s = str(value)
    s = s.replace("\\", "\\\\")
    s = s.replace("\t", "\\t")
    s = s.replace("\n", "\\n")
    s = s.replace("\r", "\\r")
    return s


_bytea_map: Optional[Dict[int, str]] = None


def quote_bytea_raw(s: Optional[bytes]) -> Optional[str]:
    """Quoting for bytea parser.  Returns None as None.

    Python implementation.
    """
    global _bytea_map
    if s is None:
        return None
    if not isinstance(s, bytes):
        raise TypeError("Expect bytes")
    if _bytea_map is None:
        _bytea_map = {}
        for i in range(256):
            c = i
            if i < 0x20 or i >= 0x7F:
                _bytea_map[c] = "\\%03o" % i
            elif i == ord("\\"):
                _bytea_map[c] = "\\\\"
            else:
                _bytea_map[c] = '%c' % i
    return "".join([_bytea_map[b] for b in s])


#
# Database specific urlencode and urldecode.
#

def db_urlencode(dict_val: Mapping[str, Any]) -> str:
    """Database specific urlencode.

    Encode None as key without '='.  That means that in "foo&bar=",
    foo is NULL and bar is empty string.

    Python implementation.
    """

    elem_list = []
    for k, v in dict_val.items():
        if v is None:
            elem = quote_plus(str(k))
        else:
            elem = quote_plus(str(k)) + '=' + quote_plus(str(v))
        elem_list.append(elem)
    return '&'.join(elem_list)


def db_urldecode(qs: str) -> Dict[str, Optional[str]]:
    """Database specific urldecode.

    Decode key without '=' as None.
    This also does not support one key several times.

    Python implementation.
    """

    res: Dict[str, Optional[str]] = {}
    for elem in qs.split('&'):
        if not elem:
            continue
        pair = elem.split('=', 1)
        name = unquote_plus(pair[0])
        if len(pair) == 1:
            res[name] = None
        else:
            res[name] = unquote_plus(pair[1])
    return res


#
# Remove C-like backslash escapes
#

_esc_re = r"\\([0-7]{1,3}|.)"
_esc_rc = re.compile(_esc_re)
_esc_map = {
    't': '\t',
    'n': '\n',
    'r': '\r',
    'a': '\a',
    'b': '\b',
    "'": "'",
    '"': '"',
    '\\': '\\',
}


def _sub_unescape_c(m: Match[str]) -> str:
    """unescape single escape seq."""
    v = m.group(1)
    if (len(v) == 1) and (v < '0' or v > '7'):
        try:
            return _esc_map[v]
        except KeyError:
            return v
    else:
        return chr(int(v, 8))


def unescape(val: str) -> str:
    """Removes C-style escapes from string.
    Python implementation.
    """
    return _esc_rc.sub(_sub_unescape_c, val)


_esql_re = r"''|\\([0-7]{1,3}|.)"
_esql_rc = re.compile(_esql_re)


def _sub_unescape_sqlext(m: Match[str]) -> str:
    """Unescape extended-quoted string."""
    if m.group() == "''":
        return "'"
    v = m.group(1)
    if (len(v) == 1) and (v < '0' or v > '7'):
        try:
            return _esc_map[v]
        except KeyError:
            return v
    return chr(int(v, 8))


def unquote_literal(val: str, stdstr: bool = False) -> Optional[str]:
    """Unquotes SQL string.

    E'..' -> extended quoting.
    '..' -> standard or extended quoting
    null -> None
    other -> returned as-is
    """
    if val[0] == "'" and val[-1] == "'":
        if stdstr:
            return val[1:-1].replace("''", "'")
        else:
            return _esql_rc.sub(_sub_unescape_sqlext, val[1:-1])
    elif len(val) > 2 and val[0] in ('E', 'e') and val[1] == "'" and val[-1] == "'":
        return _esql_rc.sub(_sub_unescape_sqlext, val[2:-1])
    elif len(val) >= 2 and val[0] == '$' and val[-1] == '$':
        p1 = val.find('$', 1)
        p2 = val.rfind('$', 1, -1)
        if p1 > 0 and p2 > p1:
            t1 = val[:p1 + 1]
            t2 = val[p2:]
            if t1 == t2:
                return val[len(t1):-len(t1)]
        raise ValueError("Bad dollar-quoted string")
    elif val.lower() == "null":
        return None
    return val




"""Helper classes for complex query generation.

Main target is code execution under PL/Python.

Query parameters are referenced as C{{key}} or C{{key:type}}.
Type will be given to C{plpy.prepare}.
If C{type} is missing, C{text} is assumed.

See L{plpy_exec} for examples.
"""

import json
import re
from functools import lru_cache
from typing import Any, Dict, List, Mapping, Optional, Sequence, Union, Tuple, cast

import skytools

from .basetypes import Cursor

try:
    import plpy
except ImportError:
    plpy = None

__all__ = (
    'QueryBuilder', 'PLPyQueryBuilder', 'PLPyQuery', 'plpy_exec',
    "run_query", "run_query_row", "run_lookup", "run_exists",
)

PARAM_INLINE = 0  # quote_literal()
PARAM_DBAPI = 1  # %()s
PARAM_PLPY = 2   # $n

_RC_PARAM = re.compile(r"""
    \{  ( [^|{}:]* )
        (?:  : ( [^|{}:]+ ) )?
        (?: \| ( [^|{}:]+ ) )?
    ( \} )?
""", re.X)


def _inline_to_text(val: Any) -> Optional[str]:
    """Approx emulate PL/Python and Psycopg2 internal conversions
    for common types.
    """
    if val is None or isinstance(val, str):
        return val
    if isinstance(val, dict):
        return json.dumps(val)
    if isinstance(val, (tuple, list)):
        return skytools.make_pgarray(val)
    if isinstance(val, bytes):
        return "\\x" + val.hex()
    return str(val)


class QArgConf:
    """Per-query arg-type config object."""
    param_type = PARAM_INLINE


class QArg:
    """Place-holder for a query parameter."""
    def __init__(self, name: str, value: Any, pos: int, conf: QArgConf):
        self.name = name
        self.value = value
        self.pos = pos
        self.conf = conf
    def __str__(self) -> str:
        if self.conf.param_type == PARAM_INLINE:
            return skytools.quote_literal(_inline_to_text(self.value))
        elif self.conf.param_type == PARAM_DBAPI:
            return "%s"
        elif self.conf.param_type == PARAM_PLPY:
            return "$%d" % self.pos
        else:
            raise Exception("bad QArgConf.param_type")


class PlanCache:
    """Cache for limited amount of plans."""

    def __init__(self, maxplans: int = 100) -> None:
        self.maxplans = maxplans

        @lru_cache(maxplans)
        def _cached_prepare(key: Tuple[str, Tuple[str, ...]]) -> Any:
            sql, types = key
            return plpy.prepare(sql, types)

        self._cached_prepare = _cached_prepare

    def get_plan(self, sql: str, types: Sequence[str]) -> Any:
        """Prepare the plan and cache it."""
        key = (sql, tuple(types))
        return self._cached_prepare(key)


class QueryBuilderCore:
    """Helper for query building.
    """

    _params: Optional[Mapping[str, Any]]
    _arg_type_list: List[str]
    _arg_value_list: List[Any]
    _sql_parts: List[Union[str, QArg]]
    _arg_conf: QArgConf
    _nargs: int

    def __init__(self, sqlexpr: str, params: Optional[Mapping[str, Any]]):
        """Init the object.

        @param sqlexpr:     Partial sql fragment.
        @param params:      Dict of parameter values.
        """
        self._params = params
        self._arg_type_list = []
        self._arg_value_list = []
        self._sql_parts = []
        self._arg_conf = QArgConf()
        self._nargs = 0

        if sqlexpr:
            self.add(sqlexpr, required=True)

    def add(self, expr: str, sql_type: str = "text", required: bool = False) -> None:
        """Add SQL fragment to query.
        """
        self._add_expr('', expr, self._params, sql_type, required)

    def get_sql(self, param_type: int = PARAM_INLINE) -> str:
        """Return generated SQL (thus far) as string.

        Possible values for param_type:
            - 0: Insert values quoted with quote_literal()
            - 1: Insert %()s in place of parameters.
            - 2: Insert $n in place of parameters.
        """
        self._arg_conf.param_type = param_type
        tmp = [str(part) for part in self._sql_parts]
        return "".join(tmp)

    def _add_expr(self, pfx: str, expr: str, params: Optional[Mapping[str, Any]], sql_type: str, required: bool) -> None:
        parts: List[Union[str, QArg]] = []
        types: List[str] = []
        values: List[Any] = []
        nargs = self._nargs
        if pfx:
            parts.append(pfx)
        pos = 0
        while True:
            # find next argument
            m = _RC_PARAM.search(expr, pos)
            if not m:
                parts.append(expr[pos:])
                break

            # add plain sql
            parts.append(expr[pos:m.start()])
            pos = m.end()

            # get arg name and type
            kparam, ktype, alt_frag, tag = m.groups()
            if not kparam or not tag:
                raise ValueError("invalid tag syntax: <%s>" % m.group(0))
            if not ktype:
                ktype = sql_type

            # params==None means params are checked later
            if params is None:
                if alt_frag is not None:
                    raise ValueError("alt_frag not supported with params=None")
            elif kparam not in params:
                if alt_frag is not None:
                    parts.append(alt_frag)
                    continue
                elif required:
                    raise Exception("required parameter missing: " + kparam)
                # optional fragment, param missing, skip it
                return

            # got arg
            nargs += 1
            if params is not None:
                val = params[kparam]
            else:
                val = kparam
            values.append(val)
            types.append(ktype)
            arg = QArg(kparam, val, nargs, self._arg_conf)
            parts.append(arg)

        # add interesting parts to the main sql
        self._sql_parts.extend(parts)
        if types:
            self._arg_type_list.extend(types)
        if values:
            self._arg_value_list.extend(values)
        self._nargs = nargs


class QueryBuilder(QueryBuilderCore):
    def execute(self, curs: Cursor) -> None:
        """Client-side query execution on DB-API 2.0 cursor.

        Calls C{curs.execute()} with proper arguments.

        Does not return anything, curs.fetch* methods
        must be called to get result.
        """
        q = self.get_sql(PARAM_DBAPI)
        args = self._params
        curs.execute(q, args)


class PLPyQueryBuilder(QueryBuilderCore):

    def __init__(self, sqlexpr: str, params: Optional[Mapping[str, Any]],
                 plan_cache: Optional[Dict[str, Any]] = None,
                 sqls: Optional[List[Dict[str, str]]] = None):
        """Init the object.

        @param sqlexpr:     Partial sql fragment.
        @param params:      Dict of parameter values.
        @param plan_cache:  (PL/Python) A dict object where to store the plan cache, under the key C{"plan_cache"}.
                            If not given, plan will not be cached and values will be inserted directly
                            to query.  Usually either C{GD} or C{SD} should be given here.
        @param sqls:        list object where to append executed sqls (used for debugging)
        """
        super().__init__(sqlexpr, params)
        self._sqls = sqls

        if plan_cache is not None:
            if 'plan_cache' not in plan_cache:
                plan_cache['plan_cache'] = PlanCache()
            self._plan_cache = plan_cache['plan_cache']
        else:
            self._plan_cache = None

    def execute(self) -> List[skytools.dbdict]:
        """Server-side query execution via plpy.

        Query can be run either cached or uncached, depending
        on C{plan_cache} setting given to L{__init__}.

        Returns result of plpy.execute().
        """

        args = self._arg_value_list
        types = self._arg_type_list

        if self._sqls is not None:
            self._sqls.append({"sql": self.get_sql(PARAM_INLINE)})

        sql = self.get_sql(PARAM_PLPY)
        if self._plan_cache is not None:
            plan = self._plan_cache.get_plan(sql, types)
        else:
            plan = plpy.prepare(sql, types)

        res = plpy.execute(plan, args)
        if res:
            return [skytools.dbdict(r) for r in res]
        else:
            return []


class PLPyQuery:
    """Static, cached PL/Python query that uses QueryBuilder formatting.

    See L{plpy_exec} for simple usage.
    """
    def __init__(self, sql: str) -> None:
        qb = QueryBuilder(sql, None)
        p_sql = qb.get_sql(PARAM_PLPY)
        p_types = qb._arg_type_list
        self.plan = plpy.prepare(p_sql, p_types)
        self.arg_map = qb._arg_value_list
        self.sql = sql

    def execute(self, arg_dict: Optional[Mapping[str, Any]], all_keys_required: bool = True) -> List[skytools.dbdict]:
        if arg_dict is None:
            arg_dict = {}
        try:
            if all_keys_required:
                arg_list = [arg_dict[k] for k in self.arg_map]
            else:
                arg_list = [arg_dict.get(k) for k in self.arg_map]
            res = plpy.execute(self.plan, arg_list)
            if res:
                return [skytools.dbdict(row) for row in res]
            return []
        except KeyError:
            need = set(self.arg_map)
            got = set(arg_dict.keys())
            missing = list(need.difference(got))
            plpy.error("Missing arguments: [%s]  QUERY: %s" % (
                ','.join(missing), repr(self.sql)))
            raise ValueError("unreachable") from None

    def __repr__(self) -> str:
        return 'PLPyQuery<%s>' % self.sql


def plpy_exec(gd: Optional[Dict[str, Any]],
              sql: str,
              args: Optional[Mapping[str, Any]],
              all_keys_required: bool = True) -> List[skytools.dbdict]:
    """Cached plan execution for PL/Python.

    @param gd:  dict to store cached plans under.  If None, caching is disabled.
    @param sql: SQL statement to execute.
    @param args: dict of arguments to query.
    @param all_keys_required: if False, missing key is taken as NULL, instead of throwing error.
    """

    if gd is None:
        return PLPyQueryBuilder(sql, args).execute()

    if 'plq_cache' not in gd:
        gd['plq_cache'] = {}
    cache = cast(Dict[str, PLPyQuery], gd['plq_cache'])

    try:
        sq = cache[sql]
    except KeyError:
        sq = PLPyQuery(sql)
        cache[sql] = sq
    return sq.execute(args, all_keys_required)


# some helper functions for convenient sql execution

def run_query(cur: Cursor, sql: str,
              params: Optional[Mapping[str, Any]] = None,
              **kwargs: Any
              ) -> List[skytools.dbdict]:
    """ Helper function if everything you need is just paramertisized execute
        Sets rows_found that is coneninet to use when you don't need result just
        want to know how many rows were affected
    """
    params = params or kwargs
    sql = QueryBuilder(sql, params).get_sql(0)
    cur.execute(sql)
    rows = cur.fetchall()
    # convert result rows to dbdict
    if rows:
        return [skytools.dbdict(r) for r in rows]
    return []


def run_query_row(cur: Cursor, sql: str, params: Optional[Mapping[str, Any]] = None, **kwargs: Any
                  ) -> Optional[skytools.dbdict]:
    """ Helper function if everything you need is just paramertisized execute to
        fetch one row only. If not found none is returned
    """
    params = params or kwargs
    rows = run_query(cur, sql, params)
    if len(rows) == 0:
        return None
    return rows[0]


def run_lookup(cur: Cursor, sql: str, params: Optional[Mapping[str, Any]] = None, **kwargs: Any) -> Any:
    """ Helper function to fetch one value Takes away all the hassle of preparing statements
        and processing returned result giving out just one value.
    """
    params = params or kwargs
    sql = QueryBuilder(sql, params).get_sql(0)
    cur.execute(sql)
    row = cur.fetchone()
    if row is None:
        return None
    return row[0]


def run_exists(cur: Cursor, sql: str, params: Optional[Mapping[str, Any]] = None, **kwargs: Any) -> bool:
    """ Helper function to fetch one value Takes away all the hassle of preparing statements
        and processing returned result giving out just one value.
    """
    params = params or kwargs
    val = run_lookup(cur, sql, params)
    return val is not None


# fake plpy for testing
class fake_plpy:
    log: List[str] = []
    def prepare(self, sql: str, types: Sequence[str]) -> Tuple[str, str, Sequence[str]]:
        self.log.append("DBG: plpy.prepare(%s, %s)" % (repr(sql), repr(types)))
        return ('PLAN', sql, types)

    def execute(self, plan: Any, args: Any = ()) -> List[skytools.dbdict]:
        self.log.append("DBG: plpy.execute(%s, %s)" % (repr(plan), repr(args)))
        return []

    def error(self, msg: str) -> None:
        self.log.append("DBG: plpy.error(%s)" % repr(msg))
        raise Exception("plpy.error")


# make plpy available
if not plpy:
    plpy = fake_plpy()
    GD: Dict[str, Any] = {}





import logging

import skytools
from skytools import skylog


def test_trace_setup() -> None:
    assert skylog.TRACE < logging.DEBUG
    assert skylog.TRACE == logging.TRACE    # type: ignore
    assert logging.getLevelName(skylog.TRACE) == "TRACE"


def test_skylog() -> None:
    log = skytools.getLogger("test.skylog")
    log.trace("tracemsg")

    assert not log.isEnabledFor(logging.TRACE)  # type: ignore





from datetime import datetime

import pytest

from skytools.timeutil import UTC, datetime_to_timestamp, parse_iso_timestamp


def test_parse_iso_timestamp() -> None:
    res = str(parse_iso_timestamp('2005-06-01 15:00'))
    assert res == '2005-06-01 15:00:00'

    res = str(parse_iso_timestamp(' 2005-06-01T15:00 +02 '))
    assert res == '2005-06-01 15:00:00+02:00'

    res = str(parse_iso_timestamp('2005-06-01 15:00:33+02:00'))
    assert res == '2005-06-01 15:00:33+02:00'

    d = parse_iso_timestamp('2005-06-01 15:00:59.33 +02')
    assert d.strftime("%z %Z") == '+0200 +02'
    assert str(parse_iso_timestamp(str(d))) == '2005-06-01 15:00:59.330000+02:00'

    res = parse_iso_timestamp('2005-06-01 15:00-0530').strftime('%Y-%m-%d %H:%M %z %Z')
    assert res == '2005-06-01 15:00 -0530 -05:30'

    res = parse_iso_timestamp('2014-10-27T11:59:13Z').strftime('%Y-%m-%d %H:%M:%S %z %Z')
    assert res == '2014-10-27 11:59:13 +0000 +00'

    with pytest.raises(ValueError):
        parse_iso_timestamp('2014.10.27')


def test_datetime_to_timestamp() -> None:
    res = datetime_to_timestamp(parse_iso_timestamp("2005-06-01 15:00:59.5 +02"))
    assert res == 1117630859.5

    res = datetime_to_timestamp(datetime.fromtimestamp(1117630859.5, UTC))
    assert res == 1117630859.5

    res = datetime_to_timestamp(datetime.fromtimestamp(1117630859.5))
    assert res == 1117630859.5

    now = datetime.utcnow()
    now2 = datetime.utcfromtimestamp(datetime_to_timestamp(now, False))
    assert abs(now2.microsecond - now.microsecond) < 100

    now2 = now2.replace(microsecond=now.microsecond)
    assert now == now2

    now = datetime.now()
    now2 = datetime.fromtimestamp(datetime_to_timestamp(now))
    assert abs(now2.microsecond - now.microsecond) < 100

    now2 = now2.replace(microsecond=now.microsecond)
    assert now == now2




"""Atomic append of gzipped data.

The point is - if several gzip streams are concatenated,
they are read back as one whole stream.
"""

import gzip
import io

__all__ = ('gzip_append',)


def gzip_append(filename: str, data: bytes, level: int = 6) -> None:
    """Append a block of data to file with safety checks."""

    # compress data
    buf = io.BytesIO()
    with gzip.GzipFile(fileobj=buf, compresslevel=level, mode="w") as g:
        g.write(data)
    zdata = buf.getvalue()

    # append, safely
    with open(filename, "ab+", 0) as f:
        f.seek(0, 2)
        pos = f.tell()
        try:
            f.write(zdata)
        except Exception as ex:
            # rollback on error
            f.seek(pos, 0)
            f.truncate()
            raise ex




"""Catch moment when tables are in sync on master and slave.
"""

import logging
import os
import subprocess
import sys
import time
from typing import IO, List, Optional, Sequence, Tuple, Dict, cast, Mapping, Any

import skytools

from .basetypes import Connection, Cursor

DictRow = Dict[str, str]


class TableRepair:
    """Checks that tables in two databases are in sync."""

    table_name: str
    fq_table_name: str
    log: logging.Logger
    pkey_list: List[str]
    common_fields: List[str]
    apply_fixes: bool
    apply_cursor: Optional[Cursor]
    cnt_insert: int
    cnt_update: int
    cnt_delete: int
    total_src: int
    total_dst: int

    def __init__(self, table_name: str, log: logging.Logger) -> None:
        self.table_name = table_name
        self.fq_table_name = skytools.quote_fqident(table_name)
        self.log = log

        self.pkey_list = []
        self.common_fields = []
        self.apply_fixes = False
        self.apply_cursor = None

        self.reset()

    def reset(self) -> None:
        self.cnt_insert = 0
        self.cnt_update = 0
        self.cnt_delete = 0
        self.total_src = 0
        self.total_dst = 0
        self.pkey_list = []
        self.common_fields = []
        self.apply_fixes = False
        self.apply_cursor = None

    def do_repair(self, src_db: Connection, dst_db: Connection, where: str,
                  pfx: str = 'repair', apply_fixes: bool = False) -> None:
        """Actual comparison."""

        self.reset()

        src_curs = src_db.cursor()
        dst_curs = dst_db.cursor()

        self.apply_fixes = apply_fixes
        if apply_fixes:
            self.apply_cursor = dst_curs

        self.log.info('Checking %s', self.table_name)

        copy_tbl = self.gen_copy_tbl(src_curs, dst_curs, where)

        dump_src = "%s.%s.src" % (pfx, self.table_name)
        dump_dst = "%s.%s.dst" % (pfx, self.table_name)
        fix = "%s.%s.fix" % (pfx, self.table_name)

        self.log.info("Dumping src table: %s", self.table_name)
        self.dump_table(copy_tbl, src_curs, dump_src)
        src_db.commit()
        self.log.info("Dumping dst table: %s", self.table_name)
        self.dump_table(copy_tbl, dst_curs, dump_dst)
        dst_db.commit()

        self.log.info("Sorting src table: %s", self.table_name)
        self.do_sort(dump_src, dump_src + '.sorted')

        self.log.info("Sorting dst table: %s", self.table_name)
        self.do_sort(dump_dst, dump_dst + '.sorted')

        self.dump_compare(dump_src + ".sorted", dump_dst + ".sorted", fix)

        os.unlink(dump_src)
        os.unlink(dump_dst)
        os.unlink(dump_src + ".sorted")
        os.unlink(dump_dst + ".sorted")

        if apply_fixes:
            dst_db.commit()

    def do_sort(self, src: str, dst: str) -> None:
        with subprocess.Popen(["sort", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE) as p:
            s_ver = p.communicate()[0]

        xenv = os.environ.copy()
        xenv['LANG'] = 'C'
        xenv['LC_ALL'] = 'C'

        cmdline = ['sort', '-T', '.']
        if s_ver.find(b"coreutils") > 0:
            cmdline.append('-S')
            cmdline.append('30%')
        cmdline.append('-o')
        cmdline.append(dst)
        cmdline.append(src)
        with subprocess.Popen(cmdline, env=xenv) as p:
            if p.wait() != 0:
                raise Exception('sort failed')

    def gen_copy_tbl(self, src_curs: Cursor, dst_curs: Cursor, where: str) -> str:
        """Create COPY expession from common fields."""
        self.pkey_list = skytools.get_table_pkeys(src_curs, self.table_name)
        dst_pkey = skytools.get_table_pkeys(dst_curs, self.table_name)
        if dst_pkey != self.pkey_list:
            self.log.error('pkeys do not match')
            sys.exit(1)

        src_cols = skytools.get_table_columns(src_curs, self.table_name)
        dst_cols = skytools.get_table_columns(dst_curs, self.table_name)
        field_list = []
        for f in self.pkey_list:
            field_list.append(f)
        for f in src_cols:
            if f in self.pkey_list:
                continue
            if f in dst_cols:
                field_list.append(f)

        self.common_fields = field_list

        fqlist = [skytools.quote_ident(col) for col in field_list]

        tbl_expr = "select %s from %s" % (",".join(fqlist), self.fq_table_name)
        if where:
            tbl_expr += ' where ' + where
        tbl_expr = "COPY (%s) TO STDOUT" % tbl_expr

        self.log.debug("using copy expr: %s", tbl_expr)

        return tbl_expr

    def dump_table(self, copy_cmd: str, curs: Cursor, fn: str) -> None:
        """Dump table to disk."""
        with open(fn, "w", 64 * 1024, encoding="utf8") as f:
            curs.copy_expert(copy_cmd, f)
            self.log.info('%s: Got %d bytes', self.table_name, f.tell())

    def get_row(self, ln: str) -> Optional[DictRow]:
        """Parse a row into dict."""
        if not ln:
            return None
        t = ln[:-1].split('\t')
        row: DictRow = {}
        for i, n in enumerate(self.common_fields):
            row[n] = t[i]
        return row

    def dump_compare(self, src_fn: str, dst_fn: str, fix: str) -> None:
        """Dump + compare single table."""
        self.log.info("Comparing dumps: %s", self.table_name)
        with open(src_fn, "r", 64 * 1024, encoding="utf8") as f1:
            with open(dst_fn, "r", 64 * 1024, encoding="utf8") as f2:
                self.dump_compare_streams(f1, f2, fix)

    def dump_compare_streams(self, f1: IO[str], f2: IO[str], fix: str) -> None:
        src_ln = f1.readline()
        dst_ln = f2.readline()
        if src_ln:
            self.total_src += 1
        if dst_ln:
            self.total_dst += 1

        if os.path.isfile(fix):
            os.unlink(fix)

        while src_ln or dst_ln:
            keep_src = keep_dst = 0
            if src_ln != dst_ln:
                src_row = self.get_row(src_ln)
                dst_row = self.get_row(dst_ln)

                diff = self.cmp_keys(src_row, dst_row)
                if diff > 0 and dst_row:
                    # src > dst
                    self.got_missed_delete(dst_row, fix)
                    keep_src = 1
                elif diff < 0 and src_row:
                    # src < dst
                    self.got_missed_insert(src_row, fix)
                    keep_dst = 1
                elif src_row and dst_row:
                    if self.cmp_data(src_row, dst_row) != 0:
                        self.got_missed_update(src_row, dst_row, fix)

            if not keep_src:
                src_ln = f1.readline()
                if src_ln:
                    self.total_src += 1
            if not keep_dst:
                dst_ln = f2.readline()
                if dst_ln:
                    self.total_dst += 1

        self.log.info("finished %s: src: %d rows, dst: %d rows,"
                      " missed: %d inserts, %d updates, %d deletes",
                      self.table_name, self.total_src, self.total_dst,
                      self.cnt_insert, self.cnt_update, self.cnt_delete)
        f1.close()
        f2.close()

    def got_missed_insert(self, src_row: DictRow, fn: str) -> None:
        """Create sql for missed insert."""
        self.cnt_insert += 1
        fld_list = self.common_fields
        fq_list = []
        val_list = []
        for f in fld_list:
            fq_list.append(skytools.quote_ident(f))
            v = skytools.unescape_copy(src_row[f])
            val_list.append(skytools.quote_literal(v))
        q = "insert into %s (%s) values (%s);" % (
            self.fq_table_name, ", ".join(fq_list), ", ".join(val_list))
        self.show_fix(q, 'insert', fn)

    def got_missed_update(self, src_row: DictRow, dst_row: DictRow, fn: str) -> None:
        """Create sql for missed update."""
        self.cnt_update += 1
        fld_list = self.common_fields
        set_list: List[str] = []
        whe_list: List[str] = []
        for f in self.pkey_list:
            self.addcmp(whe_list, skytools.quote_ident(f), skytools.unescape_copy(src_row[f]))
        for f in fld_list:
            v1 = src_row[f]
            v2 = dst_row[f]
            if self.cmp_value(v1, v2) == 0:
                continue

            self.addeq(set_list, skytools.quote_ident(f), skytools.unescape_copy(v1))
            self.addcmp(whe_list, skytools.quote_ident(f), skytools.unescape_copy(v2))

        q = "update only %s set %s where %s;" % (
            self.fq_table_name, ", ".join(set_list), " and ".join(whe_list))
        self.show_fix(q, 'update', fn)

    def got_missed_delete(self, dst_row: DictRow, fn: str) -> None:
        """Create sql for missed delete."""
        self.cnt_delete += 1
        whe_list: List[str] = []
        for f in self.pkey_list:
            self.addcmp(whe_list, skytools.quote_ident(f), skytools.unescape_copy(dst_row[f]))
        q = "delete from only %s where %s;" % (self.fq_table_name, " and ".join(whe_list))
        self.show_fix(q, 'delete', fn)

    def show_fix(self, q: str, desc: str, fn: str) -> None:
        """Print/write/apply repair sql."""
        self.log.debug("missed %s: %s", desc, q)
        with open(fn, "a", encoding="utf8") as f:
            f.write("%s\n" % q)

        if self.apply_fixes and self.apply_cursor:
            self.apply_cursor.execute(q)

    def addeq(self, dst_list: List[str], f: str, v: Optional[str]) -> None:
        """Add quoted SET."""
        vq = skytools.quote_literal(v)
        s = "%s = %s" % (f, vq)
        dst_list.append(s)

    def addcmp(self, dst_list: List[str], f: str, v: Optional[str]) -> None:
        """Add quoted comparison."""
        if v is None:
            s = "%s is null" % f
        else:
            vq = skytools.quote_literal(v)
            s = "%s = %s" % (f, vq)
        dst_list.append(s)

    def cmp_data(self, src_row: DictRow, dst_row: DictRow) -> int:
        """Compare data field-by-field."""
        for k in self.common_fields:
            v1 = src_row[k]
            v2 = dst_row[k]
            if self.cmp_value(v1, v2) != 0:
                return -1
        return 0

    def cmp_value(self, v1: str, v2: str) -> int:
        """Compare single field, tolerates tz vs notz dates."""
        if v1 == v2:
            return 0

        # try to work around tz vs. notz
        z1 = len(v1)
        z2 = len(v2)
        if z1 == z2 + 3 and z2 >= 19 and v1[z2] == '+':
            v1 = v1[:-3]
            if v1 == v2:
                return 0
        elif z1 + 3 == z2 and z1 >= 19 and v2[z1] == '+':
            v2 = v2[:-3]
            if v1 == v2:
                return 0

        return -1

    def cmp_keys(self, src_row: Optional[DictRow], dst_row: Optional[DictRow]) -> int:
        """Compare primary keys of the rows.

        Returns 1 if src > dst, -1 if src < dst and 0 if src == dst"""

        # None means table is done.  tag it larger than any existing row.
        if src_row is None:
            if dst_row is None:
                return 0
            return 1
        elif dst_row is None:
            return -1

        for k in self.pkey_list:
            v1 = src_row[k]
            v2 = dst_row[k]
            if v1 < v2:
                return -1
            elif v1 > v2:
                return 1
        return 0


class Syncer(skytools.DBScript):
    """Checks that tables in two databases are in sync."""
    lock_timeout: float = 10
    ticker_lag_limit: int = 20
    consumer_lag_limit: int = 20

    def sync_table(self, cstr1: str, cstr2: str, queue_name: str,
                   consumer_name: str, table_name: str) -> Tuple[Connection, Connection]:
        """Syncer main function.

        Returns (src_db, dst_db) that are in transaction
        where table should be in sync.
        """

        setup_db = self.get_database('setup_db', connstr=cstr1, autocommit=1)
        lock_db = self.get_database('lock_db', connstr=cstr1)

        src_db = self.get_database('src_db', connstr=cstr1, isolation_level=skytools.I_REPEATABLE_READ)
        dst_db = self.get_database('dst_db', connstr=cstr2, isolation_level=skytools.I_REPEATABLE_READ)

        lock_curs = lock_db.cursor()
        setup_curs = setup_db.cursor()
        src_curs = src_db.cursor()
        dst_curs = dst_db.cursor()

        self.check_consumer(setup_curs, queue_name, consumer_name)

        # lock table in separate connection
        self.log.info('Locking %s', table_name)
        self.set_lock_timeout(lock_curs)
        lock_time = time.time()
        lock_curs.execute("LOCK TABLE %s IN SHARE MODE" % skytools.quote_fqident(table_name))

        # now wait until consumer has updated target table until locking
        self.log.info('Syncing %s', table_name)

        # consumer must get further than this tick
        self.force_tick(setup_curs, queue_name)

        # try to force second tick also
        self.force_tick(setup_curs, queue_name)

        # take server time
        setup_curs.execute("select to_char(now(), 'YYYY-MM-DD HH24:MI:SS.MS')")
        tpos = setup_curs.fetchone()[0]
        # now wait
        while True:
            time.sleep(0.5)

            q = "select now() - lag > timestamp %s, now(), lag from pgq.get_consumer_info(%s, %s)"
            setup_curs.execute(q, [tpos, queue_name, consumer_name])
            res = setup_curs.fetchall()
            if len(res) == 0:
                raise Exception('No such consumer: %s/%s' % (queue_name, consumer_name))
            row = res[0]
            self.log.debug("tpos=%s now=%s lag=%s ok=%s", tpos, row[1], row[2], row[0])
            if row[0]:
                break

            # limit lock time
            if time.time() > lock_time + self.lock_timeout:
                self.log.error('Consumer lagging too much, exiting')
                lock_db.rollback()
                sys.exit(1)

        # take snapshot on provider side
        src_db.commit()
        src_curs.execute("SELECT 1")

        # take snapshot on subscriber side
        dst_db.commit()
        dst_curs.execute("SELECT 1")

        # release lock
        lock_db.commit()

        self.close_database('setup_db')
        self.close_database('lock_db')

        return (src_db, dst_db)

    def set_lock_timeout(self, curs: Cursor) -> None:
        ms = int(1000 * self.lock_timeout)
        if ms > 0:
            q = "SET LOCAL statement_timeout = %d" % ms
            self.log.debug(q)
            curs.execute(q)

    def check_consumer(self, curs: Cursor, queue_name: str, consumer_name: str) -> None:
        """ Before locking anything check if consumer is working ok.
        """
        self.log.info("Queue: %s Consumer: %s", queue_name, consumer_name)

        curs.execute('select current_database()')
        self.log.info('Actual db: %s', curs.fetchone()[0])

        # get ticker lag
        q = "select extract(epoch from ticker_lag) from pgq.get_queue_info(%s);"
        curs.execute(q, [queue_name])
        ticker_lag = curs.fetchone()[0]
        self.log.info("Ticker lag: %s", ticker_lag)
        # get consumer lag
        q = "select extract(epoch from lag) from pgq.get_consumer_info(%s, %s);"
        curs.execute(q, [queue_name, consumer_name])
        res = curs.fetchall()
        if len(res) == 0:
            self.log.error('check_consumer: No such consumer: %s/%s', queue_name, consumer_name)
            sys.exit(1)
        consumer_lag = res[0][0]

        # check that lag is acceptable
        self.log.info("Consumer lag: %s", consumer_lag)
        if consumer_lag > ticker_lag + 10:
            self.log.error('Consumer lagging too much, cannot proceed')
            sys.exit(1)

    def force_tick(self, curs: Cursor, queue_name: str) -> None:
        """ Force tick into source queue so that consumer can move on faster
        """
        q = "select pgq.force_tick(%s)"
        curs.execute(q, [queue_name])
        res = curs.fetchone()
        cur_pos = res[0]

        start = time.time()
        while True:
            time.sleep(0.5)
            curs.execute(q, [queue_name])
            res = curs.fetchone()
            if res[0] != cur_pos:
                # new pos
                return res[0]

            # dont loop more than 10 secs
            dur = time.time() - start
            if dur > 10 and not self.options.force:
                raise Exception("Ticker seems dead")


class Checker(Syncer):
    """Checks that tables in two databases are in sync.

    Config options::

        ## data_checker ##
        confdb = dbname=confdb host=confdb.service

        extra_connstr = user=marko

        # one of: compare, repair, repair-apply, compare-repair-apply
        check_type = compare

        # random params used in queries
        cluster_name =
        instance_name =
        proxy_host =
        proxy_db =

        # list of tables to be compared
        table_list = foo, bar, baz

        where_expr = (hashtext(key_user_name) & %%(max_slot)s) in (%%(slots)s)

        # gets no args
        source_query =
         select h.hostname, d.db_name
           from dba.cluster c
                join dba.cluster_host ch on (ch.key_cluster = c.id_cluster)
                join conf.host h on (h.id_host = ch.key_host)
                join dba.database d on (d.key_host = ch.key_host)
          where c.db_name = '%(cluster_name)s'
            and c.instance_name = '%(instance_name)s'
            and d.mk_db_type = 'partition'
            and d.mk_db_status = 'active'
          order by d.db_name, h.hostname


        target_query =
            select db_name, hostname, slots, max_slot
              from dba.get_cross_targets(%%(hostname)s, %%(db_name)s, '%(proxy_host)s', '%(proxy_db)s')

        consumer_query =
            select q.queue_name, c.consumer_name
              from conf.host h
              join dba.database d on (d.key_host = h.id_host)
              join dba.pgq_queue q on (q.key_database = d.id_database)
              join dba.pgq_consumer c on (c.key_queue = q.id_queue)
             where h.hostname = %%(hostname)s
               and d.db_name = %%(db_name)s
               and q.queue_name like 'xm%%%%'
    """

    def __init__(self, args: Sequence[str]) -> None:
        """Checker init."""
        super().__init__('data_checker', args)
        self.set_single_loop(1)
        self.log.info('Checker starting %s', str(args))

        self.lock_timeout = self.cf.getfloat('lock_timeout', 10)

        self.table_list = self.cf.getlist('table_list')

    def work(self) -> Optional[int]:
        """Syncer main function."""

        source_query = self.cf.get('source_query')
        target_query = self.cf.get('target_query')
        consumer_query = self.cf.get('consumer_query')
        where_expr = self.cf.get('where_expr')
        extra_connstr = self.cf.get('extra_connstr')

        check = self.cf.get('check_type', 'compare')

        confdb = self.get_database('confdb', autocommit=1)
        curs = confdb.cursor()

        curs.execute(source_query)
        for src_row in curs.fetchall():
            s_host = src_row['hostname']
            s_db = src_row['db_name']

            cast_row = cast(Mapping[str, Any], src_row)
            curs.execute(consumer_query, cast_row)
            r = curs.fetchone()
            consumer_name = r['consumer_name']
            queue_name = r['queue_name']

            curs.execute(target_query, cast_row)
            for dst_row in curs.fetchall():
                d_db = dst_row['db_name']
                d_host = dst_row['hostname']

                cstr1 = "dbname=%s host=%s %s" % (s_db, s_host, extra_connstr)
                cstr2 = "dbname=%s host=%s %s" % (d_db, d_host, extra_connstr)
                where = where_expr % dst_row

                self.log.info('Source: db=%s host=%s queue=%s consumer=%s',
                              s_db, s_host, queue_name, consumer_name)
                self.log.info('Target: db=%s host=%s where=%s', d_db, d_host, where)

                for tbl in self.table_list:
                    src_db, dst_db = self.sync_table(cstr1, cstr2, queue_name, consumer_name, tbl)
                    if check == 'compare':
                        self.do_compare(tbl, src_db, dst_db, where)
                    elif check == 'repair':
                        tr = TableRepair(tbl, self.log)
                        tr.do_repair(src_db, dst_db, where, 'fix.' + tbl, False)
                    elif check == 'repair-apply':
                        tr = TableRepair(tbl, self.log)
                        tr.do_repair(src_db, dst_db, where, 'fix.' + tbl, True)
                    elif check == 'compare-repair-apply':
                        ok = self.do_compare(tbl, src_db, dst_db, where)
                        if not ok:
                            tr = TableRepair(tbl, self.log)
                            tr.do_repair(src_db, dst_db, where, 'fix.' + tbl, True)
                    else:
                        raise Exception('unknown check type')
                    self.reset()
        return None

    def do_compare(self, tbl: str, src_db: Connection, dst_db: Connection, where: str) -> bool:
        """Actual comparison."""

        src_curs = src_db.cursor()
        dst_curs = dst_db.cursor()

        self.log.info('Counting %s', tbl)

        q = "select count(1) as cnt, sum(hashtext(t.*::text)) as chksum from only _TABLE_ t where %s;" % where
        q = self.cf.get('compare_sql', q)
        q = q.replace('_TABLE_', skytools.quote_fqident(tbl))

        f = "%(cnt)d rows, checksum=%(chksum)s"
        f = self.cf.get('compare_fmt', f)

        self.log.debug("srcdb: %s", q)
        src_curs.execute(q)
        src_row = src_curs.fetchone()
        src_str = f % src_row
        self.log.info("srcdb: %s", src_str)

        self.log.debug("dstdb: %s", q)
        dst_curs.execute(q)
        dst_row = dst_curs.fetchone()
        dst_str = f % dst_row
        self.log.info("dstdb: %s", dst_str)

        src_db.commit()
        dst_db.commit()

        if src_str != dst_str:
            self.log.warning("%s: Results do not match!", tbl)
            return False
        else:
            self.log.info("%s: OK!", tbl)
            return True


if __name__ == '__main__':
    script = Checker(sys.argv[1:])
    script.start()





"""Tools for Python database scripts."""

# pylint:disable=redefined-builtin,unused-wildcard-import,wildcard-import


try:
    from skytools import apipkg as _apipkg
except ImportError:
    # make pylint think everything is imported immediately
    from skytools.adminscript import *
    from skytools.config import *
    from skytools.dbservice import *
    from skytools.dbstruct import *
    from skytools.fileutil import *
    from skytools.gzlog import *
    from skytools.hashtext import *
    from skytools.natsort import *
    from skytools.parsing import *
    from skytools.psycopgwrapper import *
    from skytools.querybuilder import *
    from skytools.quoting import *
    from skytools.scripting import *
    from skytools.skylog import *
    from skytools.sockutil import *
    from skytools.sqltools import *
    from skytools.timeutil import *
    from skytools.utf8 import *

_symbols = {
    # skytools.adminscript
    'AdminScript': 'skytools.adminscript:AdminScript',
    # skytools.config
    'Config': 'skytools.config:Config',
    # skytools.dbservice
    'DBService': 'skytools.dbservice:DBService',
    'ServiceContext': 'skytools.dbservice:ServiceContext',
    'TableAPI': 'skytools.dbservice:TableAPI',
    'get_record': 'skytools.dbservice:get_record',
    'get_record_list': 'skytools.dbservice:get_record_list',
    'make_record': 'skytools.dbservice:make_record',
    'make_record_array': 'skytools.dbservice:make_record_array',
    # skytools.dbstruct
    'SeqStruct': 'skytools.dbstruct:SeqStruct',
    'TableStruct': 'skytools.dbstruct:TableStruct',
    'T_ALL': 'skytools.dbstruct:T_ALL',
    'T_CONSTRAINT': 'skytools.dbstruct:T_CONSTRAINT',
    'T_DEFAULT': 'skytools.dbstruct:T_DEFAULT',
    'T_GRANT': 'skytools.dbstruct:T_GRANT',
    'T_INDEX': 'skytools.dbstruct:T_INDEX',
    'T_OWNER': 'skytools.dbstruct:T_OWNER',
    'T_PARENT': 'skytools.dbstruct:T_PARENT',
    'T_PKEY': 'skytools.dbstruct:T_PKEY',
    'T_RULE': 'skytools.dbstruct:T_RULE',
    'T_SEQUENCE': 'skytools.dbstruct:T_SEQUENCE',
    'T_TABLE': 'skytools.dbstruct:T_TABLE',
    'T_TRIGGER': 'skytools.dbstruct:T_TRIGGER',
    # skytools.fileutil
    'signal_pidfile': 'skytools.fileutil:signal_pidfile',
    'write_atomic': 'skytools.fileutil:write_atomic',
    # skytools.gzlog
    'gzip_append': 'skytools.gzlog:gzip_append',
    # skytools.hashtext
    'hashtext_old': 'skytools.hashtext:hashtext_old',
    'hashtext_new': 'skytools.hashtext:hashtext_new',
    # skytools.natsort
    'natsort': 'skytools.natsort:natsort',
    'natsort_icase': 'skytools.natsort:natsort_icase',
    'natsorted': 'skytools.natsort:natsorted',
    'natsorted_icase': 'skytools.natsort:natsorted_icase',
    'natsort_key': 'skytools.natsort:natsort_key',
    'natsort_key_icase': 'skytools.natsort:natsort_key_icase',
    # skytools.parsing
    'dedent': 'skytools.parsing:dedent',
    'hsize_to_bytes': 'skytools.parsing:hsize_to_bytes',
    'merge_connect_string': 'skytools.parsing:merge_connect_string',
    'parse_acl': 'skytools.parsing:parse_acl',
    'parse_connect_string': 'skytools.parsing:parse_connect_string',
    'parse_logtriga_sql': 'skytools.parsing:parse_logtriga_sql',
    'parse_pgarray': 'skytools.parsing:parse_pgarray',
    'parse_sqltriga_sql': 'skytools.parsing:parse_sqltriga_sql',
    'parse_statements': 'skytools.parsing:parse_statements',
    'parse_tabbed_table': 'skytools.parsing:parse_tabbed_table',
    'sql_tokenizer': 'skytools.parsing:sql_tokenizer',
    # skytools.psycopgwrapper
    'connect_database': 'skytools.psycopgwrapper:connect_database',
    'DBError': 'skytools.psycopgwrapper:DBError',
    'I_AUTOCOMMIT': 'skytools.psycopgwrapper:I_AUTOCOMMIT',
    'I_READ_COMMITTED': 'skytools.psycopgwrapper:I_READ_COMMITTED',
    'I_REPEATABLE_READ': 'skytools.psycopgwrapper:I_REPEATABLE_READ',
    'I_SERIALIZABLE': 'skytools.psycopgwrapper:I_SERIALIZABLE',
    # skytools.querybuilder
    'PLPyQuery': 'skytools.querybuilder:PLPyQuery',
    'PLPyQueryBuilder': 'skytools.querybuilder:PLPyQueryBuilder',
    'QueryBuilder': 'skytools.querybuilder:QueryBuilder',
    'plpy_exec': 'skytools.querybuilder:plpy_exec',
    'run_exists': 'skytools.querybuilder:run_exists',
    'run_lookup': 'skytools.querybuilder:run_lookup',
    'run_query': 'skytools.querybuilder:run_query',
    'run_query_row': 'skytools.querybuilder:run_query_row',
    # skytools.quoting
    'db_urldecode': 'skytools.quoting:db_urldecode',
    'db_urlencode': 'skytools.quoting:db_urlencode',
    'json_decode': 'skytools.quoting:json_decode',
    'json_encode': 'skytools.quoting:json_encode',
    'make_pgarray': 'skytools.quoting:make_pgarray',
    'quote_bytea_copy': 'skytools.quoting:quote_bytea_copy',
    'quote_bytea_literal': 'skytools.quoting:quote_bytea_literal',
    'quote_bytea_raw': 'skytools.quoting:quote_bytea_raw',
    'quote_copy': 'skytools.quoting:quote_copy',
    'quote_fqident': 'skytools.quoting:quote_fqident',
    'quote_ident': 'skytools.quoting:quote_ident',
    'quote_json': 'skytools.quoting:quote_json',
    'quote_literal': 'skytools.quoting:quote_literal',
    'quote_statement': 'skytools.quoting:quote_statement',
    'unescape': 'skytools.quoting:unescape',
    'unescape_copy': 'skytools.quoting:unescape_copy',
    'unquote_fqident': 'skytools.quoting:unquote_fqident',
    'unquote_ident': 'skytools.quoting:unquote_ident',
    'unquote_literal': 'skytools.quoting:unquote_literal',
    # skytools.scripting
    'BaseScript': 'skytools.scripting:BaseScript',
    'daemonize': 'skytools.scripting:daemonize',
    'DBScript': 'skytools.scripting:DBScript',
    'UsageError': 'skytools.scripting:UsageError',
    # skytools.skylog
    'getLogger': 'skytools.skylog:getLogger',
    # skytools.sockutil
    'set_cloexec': 'skytools.sockutil:set_cloexec',
    'set_nonblocking': 'skytools.sockutil:set_nonblocking',
    'set_tcp_keepalive': 'skytools.sockutil:set_tcp_keepalive',
    # skytools.sqltools
    'dbdict': 'skytools.sqltools:dbdict',
    'CopyPipe': 'skytools.sqltools:CopyPipe',
    'DBFunction': 'skytools.sqltools:DBFunction',
    'DBLanguage': 'skytools.sqltools:DBLanguage',
    'DBObject': 'skytools.sqltools:DBObject',
    'DBSchema': 'skytools.sqltools:DBSchema',
    'DBTable': 'skytools.sqltools:DBTable',
    'Snapshot': 'skytools.sqltools:Snapshot',
    'db_install': 'skytools.sqltools:db_install',
    'exists_function': 'skytools.sqltools:exists_function',
    'exists_language': 'skytools.sqltools:exists_language',
    'exists_schema': 'skytools.sqltools:exists_schema',
    'exists_sequence': 'skytools.sqltools:exists_sequence',
    'exists_table': 'skytools.sqltools:exists_table',
    'exists_temp_table': 'skytools.sqltools:exists_temp_table',
    'exists_type': 'skytools.sqltools:exists_type',
    'exists_view': 'skytools.sqltools:exists_view',
    'fq_name': 'skytools.sqltools:fq_name',
    'fq_name_parts': 'skytools.sqltools:fq_name_parts',
    'full_copy': 'skytools.sqltools:full_copy',
    'get_table_columns': 'skytools.sqltools:get_table_columns',
    'get_table_oid': 'skytools.sqltools:get_table_oid',
    'get_table_pkeys': 'skytools.sqltools:get_table_pkeys',
    'installer_apply_file': 'skytools.sqltools:installer_apply_file',
    'installer_find_file': 'skytools.sqltools:installer_find_file',
    'magic_insert': 'skytools.sqltools:magic_insert',
    'mk_delete_sql': 'skytools.sqltools:mk_delete_sql',
    'mk_insert_sql': 'skytools.sqltools:mk_insert_sql',
    'mk_update_sql': 'skytools.sqltools:mk_update_sql',
    # skytools.timeutil
    'FixedOffsetTimezone': 'skytools.timeutil:FixedOffsetTimezone',
    'datetime_to_timestamp': 'skytools.timeutil:datetime_to_timestamp',
    'parse_iso_timestamp': 'skytools.timeutil:parse_iso_timestamp',
    # skytools.utf8
    'safe_utf8_decode': 'skytools.utf8:safe_utf8_decode',
}

__all__ = tuple(_symbols)
_symbols['__version__'] = 'skytools.installer_config:package_version'

# lazy-import exported vars
_apipkg.initpkg(__name__, _symbols, {'apipkg': _apipkg})  # type: ignore





import pytest

from skytools.sqltools import (
    Snapshot, dbdict, fq_name, fq_name_parts, magic_insert,
    mk_delete_sql, mk_insert_sql, mk_update_sql,
)


def test_dbdict() -> None:
    row = dbdict(a=1, b=2)
    assert (row.a, row.b, row['a'], row['b']) == (1, 2, 1, 2)
    row.c = 3
    assert row['c'] == 3

    del row.c
    with pytest.raises(AttributeError):
        assert row.c

    with pytest.raises(KeyError):
        assert row['c']

    row.merge({'q': 4})
    assert row.q == 4


def test_fq_name_parts() -> None:
    assert fq_name_parts('tbl') == ('public', 'tbl')
    assert fq_name_parts('foo.tbl') == ('foo', 'tbl')
    assert fq_name_parts('foo.tbl.baz') == ('foo', 'tbl.baz')


def test_fq_name() -> None:
    assert fq_name('tbl') == 'public.tbl'
    assert fq_name('foo.tbl') == 'foo.tbl'
    assert fq_name('foo.tbl.baz') == 'foo.tbl.baz'


def test_snapshot() -> None:
    sn = Snapshot('11:20:11,12,15')
    assert sn.contains(9)
    assert not sn.contains(11)
    assert sn.contains(17)
    assert not sn.contains(20)
    with pytest.raises(ValueError):
        Snapshot(':')


def test_magic_insert() -> None:
    res = magic_insert(None, 'tbl', [[1, '1'], [2, '2']], ['col1', 'col2'])
    exp = 'COPY public.tbl (col1,col2) FROM STDIN;\n1\t1\n2\t2\n\\.\n'
    assert res == exp

    res = magic_insert(None, 'tbl', [[1, '1'], [2, '2']], ['col1', 'col2'], use_insert=True)
    exp = "insert into public.tbl (col1,col2) values ('1','1');\ninsert into public.tbl (col1,col2) values ('2','2');\n"
    assert res == exp

    assert magic_insert(None, 'tbl', [], ['col1', 'col2']) is None

    res = magic_insert(None, 'tbl."1"', [[1, '1'], [2, '2']], ['col1', 'col2'], quoted_table=True)
    exp = 'COPY tbl."1" (col1,col2) FROM STDIN;\n1\t1\n2\t2\n\\.\n'
    assert res == exp

    with pytest.raises(Exception):
        magic_insert(None, 'tbl."1"', [[1, '1'], [2, '2']])

    res = magic_insert(None, 'a.tbl', [{'a': 1}, {'a': 2}])
    exp = 'COPY a.tbl (a) FROM STDIN;\n1\n2\n\\.\n'
    assert res == exp

    res = magic_insert(None, 'a.tbl', [{'a': 1}, {'a': 2}], use_insert=True)
    exp = "insert into a.tbl (a) values ('1');\ninsert into a.tbl (a) values ('2');\n"
    assert res == exp

    # More fields than data

    res = magic_insert(None, 'tbl', [[1, 'a']], ['col1', 'col2', 'col3'])
    exp = 'COPY public.tbl (col1,col2,col3) FROM STDIN;\n1\ta\t\\N\n\\.\n'
    assert res == exp

    res = magic_insert(None, 'tbl', [[1, 'a']], ['col1', 'col2', 'col3'], use_insert=True)
    exp = "insert into public.tbl (col1,col2,col3) values ('1','a',null);\n"
    assert res == exp

    res = magic_insert(None, 'tbl', [{'a': 1}, {'b': 2}], ['a', 'b'], use_insert=False)
    exp = 'COPY public.tbl (a,b) FROM STDIN;\n1\t\\N\n\\N\t2\n\\.\n'
    assert res == exp

    res = magic_insert(None, 'tbl', [{'a': 1}, {'b': 2}], ['a', 'b'], use_insert=True)
    exp = "insert into public.tbl (a,b) values ('1',null);\ninsert into public.tbl (a,b) values (null,'2');\n"
    assert res == exp


def test_mk_insert_sql() -> None:
    row = {'id': 1, 'data': None}
    res = mk_insert_sql(row, 'tbl')
    exp = "insert into public.tbl (id, data) values ('1', null);"
    assert res == exp

    fmap = {'id': 'id_', 'data': 'data_'}
    res = mk_insert_sql(row, 'tbl', ['x'], fmap)
    exp = "insert into public.tbl (id_, data_) values ('1', null);"
    assert res == exp


def test_mk_update_sql() -> None:
    res = mk_update_sql({'id': 0, 'id2': '2', 'data': 'str\\'}, 'Table', ['id', 'id2'])
    exp = 'update only public."Table" set data = E\'str\\\\\' where id = \'0\' and id2 = \'2\';'
    assert res == exp

    res = mk_update_sql({'id': 0, 'id2': '2', 'data': 'str\\'}, 'Table', ['id', 'id2'],
                        {'id': '_id', 'id2': '_id2', 'data': '_data'})
    exp = 'update only public."Table" set _data = E\'str\\\\\' where _id = \'0\' and _id2 = \'2\';'
    assert res == exp

    with pytest.raises(Exception):
        mk_update_sql({'id': 0, 'id2': '2', 'data': 'str\\'}, 'Table', [])


def test_mk_delete_sql() -> None:
    res = mk_delete_sql({'a': 1, 'b': 2, 'c': 3}, 'tablename', ['a', 'b'])
    exp = "delete from only public.tablename where a = '1' and b = '2';"
    assert res == exp

    res = mk_delete_sql({'a': 1, 'b': 2, 'c': 3}, 'tablename', ['a', 'b'], {'a': 'aa', 'b': 'bb'})
    exp = "delete from only public.tablename where aa = '1' and bb = '2';"
    assert res == exp

    with pytest.raises(Exception):
        mk_delete_sql({'a': 1, 'b': 2, 'c': 3}, 'tablename', [])




r"""UTF-8 sanitizer.

Python's UTF-8 parser is quite relaxed, this creates problems when
talking with other software that uses stricter parsers.
"""

import codecs
import re
from typing import Match, Optional, Pattern, Tuple

__all__ = ('safe_utf8_decode', 'sanitize_unicode')

# by default, use same symbol as 'replace'
REPLACEMENT_SYMBOL = chr(0xFFFD)   # 65533

_urc: Optional[Pattern[str]] = None


def _fix_utf8(m: Match[str]) -> str:
    """Merge UTF16 surrogates, replace others"""
    u = m.group()
    if len(u) == 2:
        # merge into single symbol
        c1 = ord(u[0])
        c2 = ord(u[1])
        c = 0x10000 + ((c1 & 0x3FF) << 10) + (c2 & 0x3FF)
        return chr(c)
    else:
        # use replacement symbol
        return REPLACEMENT_SYMBOL


def sanitize_unicode(u: str) -> str:
    """Fix invalid symbols in unicode string."""
    global _urc

    if not isinstance(u, str):
        raise TypeError('Need unicode string')

    # regex for finding invalid chars, works on unicode string
    if not _urc:
        rx = "[\uD800-\uDBFF] [\uDC00-\uDFFF]? | [\0\uDC00-\uDFFF]"
        _urc = re.compile(rx, re.X)

    # now find and fix UTF16 surrogates
    m = _urc.search(u)
    if m:
        u = _urc.sub(_fix_utf8, u)
    return u


def safe_replace(exc: UnicodeError) -> Tuple[str, int]:
    """Replace only one symbol at a time.

    Builtin .decode('xxx', 'replace') replaces several symbols
    together, which is unsafe.
    """
    c2 = REPLACEMENT_SYMBOL

    # we could assume latin1
    #if 0:
    #    c1 = exc.object[exc.start]
    #    c2 = chr(ord(c1))
    assert isinstance(exc, UnicodeDecodeError)
    return c2, exc.start + 1


# register, it will be globally available
codecs.register_error("safe_replace", safe_replace)


def safe_utf8_decode(s: bytes) -> Tuple[bool, str]:
    """Decode UTF-8 safely.

    Acts like str.decode('utf8', 'replace') but also fixes
    UTF16 surrogates and NUL bytes, which Python's default
    decoder does not do.

    @param s: utf8-encoded byte string
    @return: tuple of (was_valid_utf8, unicode_string)
    """

    # decode with error detection
    ok = True
    try:
        # expect no errors by default
        u = s.decode('utf8')
    except UnicodeDecodeError:
        u = s.decode('utf8', 'safe_replace')
        ok = False

    u2 = sanitize_unicode(u)
    if u is not u2:
        ok = False
    return (ok, u2)




"""Database tools.
"""

import abc
import io
import typing
import types

from typing import (
    IO, Any, Mapping, Optional, Sequence, Tuple, Type, Union,
    KeysView, ValuesView, ItemsView, Iterator,
)

try:
    from typing import Protocol
except ImportError:
    Protocol = object   # type: ignore

__all__ = (
    "ExecuteParams", "DictRow",
    "Cursor", "Connection",
    "Runnable",
    "HasFileno", "FileDescriptor", "FileDescriptorLike",
    "Buffer",
)

ExecuteParams = Union[Sequence[Any], Mapping[str, Any]]


class DictRow(Protocol):
    """Allow both key and index-based access.

    Both Psycopg2 DictRow and PL/Python rows support this.
    """
    def keys(self) -> KeysView[str]: raise NotImplementedError
    def values(self) -> ValuesView[Any]: raise NotImplementedError
    def items(self) -> ItemsView[str, Any]: raise NotImplementedError
    def get(self, key: str, default: Any = None) -> Any: raise NotImplementedError
    def __getitem__(self, key: Union[str, int]) -> Any: raise NotImplementedError
    def __iter__(self) -> Iterator[str]: raise NotImplementedError
    def __len__(self) -> int: raise NotImplementedError
    def __contains__(self, key: object) -> bool: raise NotImplementedError


class Cursor(Protocol):
    @property
    def rowcount(self) -> int: raise NotImplementedError
    @property
    def statusmessage(self) -> Optional[str]: raise NotImplementedError
    def execute(self, sql: str, params: Optional[ExecuteParams] = None) -> None: raise NotImplementedError
    def fetchall(self) -> Sequence[DictRow]: raise NotImplementedError
    def fetchone(self) -> DictRow: raise NotImplementedError
    def __enter__(self) -> "Cursor": raise NotImplementedError
    def __exit__(self, typ: Optional[Type[BaseException]], exc: Optional[BaseException], tb: Optional[types.TracebackType]) -> None:
        raise NotImplementedError
    def copy_expert(
        self, sql: str,
        f: Union[IO[str], IO[bytes], io.TextIOBase, io.RawIOBase],
        size: int = 8192
    ) -> None:
        raise NotImplementedError
    def fileno(self) -> int: raise NotImplementedError
    @property
    def description(self) -> Sequence[Tuple[str, int, int, int, Optional[int], Optional[int], None]]: raise NotImplementedError
    @property
    def connection(self) -> "Connection": raise NotImplementedError


class Connection(Protocol):
    def cursor(self) -> Cursor: raise NotImplementedError
    def rollback(self) -> None: raise NotImplementedError
    def commit(self) -> None: raise NotImplementedError
    def close(self) -> None: raise NotImplementedError
    @property
    def isolation_level(self) -> int: raise NotImplementedError
    def set_isolation_level(self, level: int) -> None: raise NotImplementedError
    @property
    def encoding(self) -> str: raise NotImplementedError
    def set_client_encoding(self, encoding: str) -> None: raise NotImplementedError
    @property
    def server_version(self) -> int: raise NotImplementedError
    def __enter__(self) -> "Connection": raise NotImplementedError
    def __exit__(self, typ: Optional[Type[BaseException]], exc: Optional[BaseException], tb: Optional[types.TracebackType]) -> None:
        raise NotImplementedError

class Runnable(Protocol):
    def run(self) -> None: raise NotImplementedError


class HasFileno(Protocol):
    def fileno(self) -> int: raise NotImplementedError


FileDescriptor = int
FileDescriptorLike = Union[int, HasFileno]

try:
    from typing_extensions import Buffer
except ImportError:
    if typing.TYPE_CHECKING:
        from _typeshed import Buffer    # type: ignore
    else:
        try:
            from collections.abc import Buffer  # type: ignore
        except ImportError:
            class Buffer(abc.ABC):
                pass
            Buffer.register(memoryview)
            Buffer.register(bytearray)
            Buffer.register(bytes)




from skytools.hashtext import (
    hashtext_new, hashtext_new_py, hashtext_old, hashtext_old_py,
)


def test_hashtext_new_const() -> None:
    c0 = [hashtext_new_py(b'x' * (0 * 5 + j)) for j in range(5)]
    c1 = [hashtext_new_py(b'x' * (1 * 5 + j)) for j in range(5)]
    c2 = [hashtext_new_py(b'x' * (2 * 5 + j)) for j in range(5)]

    assert c0 == [-1477818771, 1074944137, -1086392228, -1992236649, -1379736791]
    assert c1 == [-370454118, 1489915569, -66683019, -2126973000, 1651296771]
    assert c2 == [755764456, -1494243903, 631527812, 28686851, -9498641]


def test_hashtext_old_const() -> None:
    c0 = [hashtext_old_py(b'x' * (0 * 5 + j)) for j in range(5)]
    c1 = [hashtext_old_py(b'x' * (1 * 5 + j)) for j in range(5)]
    c2 = [hashtext_old_py(b'x' * (2 * 5 + j)) for j in range(5)]

    assert c0 == [-863449762, 37835117, 294739542, -320432768, 1007638138]
    assert c1 == [1422906842, -261065348, 59863994, -162804943, 1736144510]
    assert c2 == [-682756517, 317827663, -495599455, -1411793989, 1739997714]


def test_hashtext_new_impl() -> None:
    data = b'HypficUjFitraxlumCitcemkiOkIkthi'
    p = [hashtext_new_py(data[:l]) for l in range(len(data) + 1)]
    c = [hashtext_new(data[:l]) for l in range(len(data) + 1)]
    assert p == c, '%s <> %s' % (p, c)


def test_hashtext_old_impl() -> None:
    data = b'HypficUjFitraxlumCitcemkiOkIkthi'
    p = [hashtext_old_py(data[:l]) for l in range(len(data) + 1)]
    c = [hashtext_old(data[:l]) for l in range(len(data) + 1)]
    assert p == c, '%s <> %s' % (p, c)


def test_hashtext_new_input_types() -> None:
    data = b'HypficUjFitraxlumCitcemkiOkIkthi'
    exp = hashtext_new(data)
    assert hashtext_new(data.decode("utf8")) == exp
    #assert hashtext_new(memoryview(data)) == exp
    #assert hashtext_new(bytearray(data)) == exp

    assert hashtext_new_py(data) == exp
    assert hashtext_new_py(data.decode("utf8")) == exp
    #assert hashtext_new_py(memoryview(data)) == exp
    #assert hashtext_new_py(bytearray(data)) == exp


def test_hashtext_old_input_types() -> None:
    data = b'HypficUjFitraxlumCitcemkiOkIkthi'
    exp = hashtext_old(data)
    assert hashtext_old(data.decode("utf8")) == exp
    #assert hashtext_old(memoryview(data)) == exp
    #assert hashtext_old(bytearray(data)) == exp

    assert hashtext_old_py(data) == exp
    assert hashtext_old_py(data.decode("utf8")) == exp
    #assert hashtext_old_py(memoryview(data)) == exp
    #assert hashtext_old_py(bytearray(data)) == exp




"""Wrapper around psycopg2.

Database connection provides regular DB-API 2.0 interface.

Connection object methods::

    .cursor()

    .commit()

    .rollback()

    .close()

Cursor methods::

    .execute(query[, args])

    .fetchone()

    .fetchall()


Sample usage::

    db = self.get_database('somedb')
    curs = db.cursor()

    # query arguments as array
    q = "select * from table where id = %s and name = %s"
    curs.execute(q, [1, 'somename'])

    # query arguments as dict
    q = "select id, name from table where id = %(id)s and name = %(name)s"
    curs.execute(q, {'id': 1, 'name': 'somename'})

    # loop over resultset
    for row in curs.fetchall():

        # columns can be asked by index:
        id = row[0]
        name = row[1]

        # and by name:
        id = row['id']
        name = row['name']

    # now commit the transaction
    db.commit()
"""
from typing import cast

import psycopg2
import psycopg2.extensions
import psycopg2.extras
from psycopg2 import Error as DBError

from .basetypes import Connection

__all__ = (
    'connect_database', 'DBError', 'I_AUTOCOMMIT', 'I_READ_COMMITTED',
    'I_REPEATABLE_READ', 'I_SERIALIZABLE',
)

#: Isolation level for db.set_isolation_level()
I_AUTOCOMMIT = psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT

#: Isolation level for db.set_isolation_level()
I_READ_COMMITTED = psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED

#: Isolation level for db.set_isolation_level()
I_REPEATABLE_READ = psycopg2.extensions.ISOLATION_LEVEL_REPEATABLE_READ

#: Isolation level for db.set_isolation_level()
I_SERIALIZABLE = psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE


def connect_database(connstr: str) -> Connection:
    """Create a db connection with DictCursor.
    """
    db = psycopg2.connect(connstr, cursor_factory=psycopg2.extras.DictCursor)
    return cast(Connection, db)





from typing import Dict, Any

import pytest

from skytools.querybuilder import ( # type: ignore[attr-defined]
    PARAM_DBAPI, PARAM_INLINE, PARAM_PLPY,
    PlanCache, QueryBuilder, plpy, plpy_exec,
)


def test_cached_plan() -> None:
    cache = PlanCache(3)

    p1 = cache.get_plan('sql1', ['text'])
    assert p1 is cache.get_plan('sql1', ['text'])

    p2 = cache.get_plan('sql1', ['int'])
    assert p2 is cache.get_plan('sql1', ['int'])
    assert p1 is not p2

    p3 = cache.get_plan('sql3', ['text'])
    assert p3 is cache.get_plan('sql3', ['text'])

    p4 = cache.get_plan('sql4', ['text'])
    assert p4 is cache.get_plan('sql4', ['text'])

    p1x = cache.get_plan('sql1', ['text'])
    assert p1 is not p1x


def test_querybuilder_core() -> None:
    args = {'success': 't', 'total': 45, 'ccy': 'EEK', 'id': 556}
    q = QueryBuilder("update orders set total = {total} where id = {id}", args)
    q.add(" and optional = {non_exist}")
    q.add(" and final = {success}")
    exp = "update orders set total = '45' where id = '556' and final = 't'"
    assert q.get_sql(PARAM_INLINE) == exp
    exp = "update orders set total = %s where id = %s and final = %s"
    assert q.get_sql(PARAM_DBAPI) == exp
    exp = "update orders set total = $1 where id = $2 and final = $3"
    assert q.get_sql(PARAM_PLPY) == exp


def test_querybuilder_parse_errors() -> None:
    args = {'id': 1}
    with pytest.raises(ValueError):
        QueryBuilder("values ({{id)", args)
    with pytest.raises(ValueError):
        QueryBuilder("values ({id)", args)
    with pytest.raises(ValueError):
        QueryBuilder("values ({id::})", args)
    with pytest.raises(ValueError):
        QueryBuilder("values ({id||})", args)


def test_querybuilder_inline() -> None:
    from decimal import Decimal
    args = {
        'list': [1, 2], 'tup': ('s', 'x'), 'dict': {'a': 1}, 'none': None,
        'bin': b'bin', 'str': 's', 'dec': Decimal('1.1'),
    }
    q = QueryBuilder("values ({list}, {dict}, {tup}, {none}, {str}, {bin}, {dec})", args)
    exp = r"""values ('{1,2}', '{"a": 1}', '{s,x}', null, 's', E'\\x62696e', '1.1')"""
    assert q.get_sql(PARAM_INLINE) == exp


def test_querybuilder_altsql() -> None:
    args = {'id': 1}
    q = QueryBuilder("values ({id|XX}, {missing|DEFAULT})", args)
    exp = "values ('1', DEFAULT)"
    assert q.get_sql(PARAM_INLINE) == exp

    with pytest.raises(ValueError):
        QueryBuilder("values ({missing|DEFAULT})", None)


def test_plpy_exec() -> None:
    GD: Dict[str, Any] = {}
    plpy.log.clear()
    plpy_exec(GD, "select {arg1}, {arg2:int4}, {arg1}", {'arg1': '1', 'arg2': '2'})
    assert plpy.log == [
        "DBG: plpy.prepare('select $1, $2, $3', ['text', 'int4', 'text'])",
        "DBG: plpy.execute(('PLAN', 'select $1, $2, $3', ['text', 'int4', 'text']), ['1', '2', '1'])",
    ]

    plpy.log.clear()
    plpy_exec(None, "select {arg1}, {arg2:int4}, {arg1}", {'arg1': '1', 'arg2': '2'})
    assert plpy.log == [
        "DBG: plpy.prepare('select $1, $2, $3', ['text', 'int4', 'text'])",
        "DBG: plpy.execute(('PLAN', 'select $1, $2, $3', ['text', 'int4', 'text']), ['1', '2', '1'])",
    ]

    plpy.log.clear()
    plpy_exec(GD, "select {arg1}, {arg2:int4}, {arg1}", {'arg1': '3', 'arg2': '4'})
    assert plpy.log == [
        "DBG: plpy.execute(('PLAN', 'select $1, $2, $3', ['text', 'int4', 'text']), ['3', '4', '3'])"
    ]

    plpy.log.clear()
    with pytest.raises(Exception):
        plpy_exec(GD, "select {arg1}, {arg2:int4}, {arg1}", {'arg1': '3'})
    assert plpy.log == [
        """DBG: plpy.error("Missing arguments: [arg2]  QUERY: 'select {arg1}, {arg2:int4}, {arg1}'")"""
    ]

    plpy.log.clear()
    plpy_exec(GD, "select {arg1}, {arg2:int4}, {arg1}", {'arg1': '3'}, False)
    assert plpy.log == [
        "DBG: plpy.execute(('PLAN', 'select $1, $2, $3', ['text', 'int4', 'text']), ['3', None, '3'])"
    ]




"""Useful functions and classes for database scripts.
"""

import argparse
import errno
import logging
import logging.config
import logging.handlers
import optparse
import os
import select
import signal
import sys
import time
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union, Callable, Type, cast

import skytools
import skytools.skylog

from .basetypes import Connection, Runnable, Cursor, DictRow, ExecuteParams

try:
    import skytools.installer_config
    default_skylog = skytools.installer_config.skylog
except ImportError:
    default_skylog = 0


__all__ = (
    'BaseScript', 'UsageError', 'daemonize', 'DBScript',
)


class UsageError(Exception):
    """User induced error."""


#
# daemon mode
#

def daemonize() -> None:
    """Turn the process into daemon.

    Goes background and disables all i/o.
    """

    # launch new process, kill parent
    pid = os.fork()
    if pid != 0:
        os._exit(0)

    # start new session
    os.setsid()

    # stop i/o
    fd = os.open("/dev/null", os.O_RDWR)
    os.dup2(fd, 0)
    os.dup2(fd, 1)
    os.dup2(fd, 2)
    if fd > 2:
        os.close(fd)


#
# Pidfile locking+cleanup & daemonization combined
#

def run_single_process(runnable: Runnable, daemon: bool, pidfile: Optional[str]) -> None:
    """Run runnable class, possibly daemonized, locked on pidfile."""

    # check if another process is running
    if pidfile and os.path.isfile(pidfile):
        if skytools.signal_pidfile(pidfile, 0):
            print("Pidfile exists, another process running?")
            sys.exit(1)
        else:
            print("Ignoring stale pidfile")

    # daemonize if needed
    if daemon:
        daemonize()

    # clean only own pidfile
    own_pidfile = False

    try:
        if pidfile:
            data = str(os.getpid())
            skytools.write_atomic(pidfile, data)
            own_pidfile = True

        runnable.run()
    finally:
        if own_pidfile and pidfile:
            try:
                os.remove(pidfile)
            except BaseException:
                pass


#
# logging setup
#

_log_config_done: int = 0
_log_init_done: Dict[str, int] = {}


def _init_log(job_name: str, service_name: str, cf: skytools.Config, log_level: int, is_daemon: bool) -> logging.Logger:
    """Logging setup happens here."""
    global _log_config_done

    got_skylog = 0
    use_skylog = cf.getint("use_skylog", default_skylog)

    # if non-daemon, avoid skylog if script is running on console.
    # set use_skylog=2 to disable.
    if not is_daemon and use_skylog == 1:
        # pylint gets spooked by it's own stdout wrapper and refuses to shut down
        # about it.  'noqa' tells prospector to ignore all warnings here.
        if sys.stdout.isatty():   # noqa
            use_skylog = 0

    # load logging config if needed
    if use_skylog and not _log_config_done:
        # python logging.config braindamage:
        # cannot specify external classess without such hack
        logging.skylog = skytools.skylog    # type: ignore
        skytools.skylog.set_service_name(service_name, job_name)

        # load general config
        flist = cf.getlist('skylog_locations',
                           ['skylog.ini', '~/.skylog.ini', '/etc/skylog.ini'])
        for fn in flist:
            fn = os.path.expanduser(fn)
            if os.path.isfile(fn):
                defs = {'job_name': job_name, 'service_name': service_name}
                logging.config.fileConfig(fn, defs, False)
                got_skylog = 1
                break
        _log_config_done = 1
        if not got_skylog:
            sys.stderr.write("skylog.ini not found!\n")
            sys.exit(1)

    # avoid duplicate logging init for job_name
    log = logging.getLogger(job_name)
    if job_name in _log_init_done:
        return log
    _log_init_done[job_name] = 1

    # tune level on root logger
    root = logging.getLogger()
    root.setLevel(log_level)

    # compatibility: specify ini file in script config
    def_fmt = '%(asctime)s %(process)s %(levelname)s %(message)s'
    def_datefmt = ''  # None
    logfile = cf.getfile("logfile", "")
    if logfile:
        fstr = cf.get('logfmt_file', def_fmt)
        fstr_date = cf.get('logdatefmt_file', def_datefmt)
        if log_level < logging.INFO:
            fstr = cf.get('logfmt_file_verbose', fstr)
            fstr_date = cf.get('logdatefmt_file_verbose', fstr_date)
        fmt = logging.Formatter(fstr, fstr_date)
        size = cf.getint('log_size', 10 * 1024 * 1024)
        num = cf.getint('log_count', 3)
        file_hdlr = logging.handlers.RotatingFileHandler(
            logfile, 'a', size, num)
        file_hdlr.setFormatter(fmt)
        root.addHandler(file_hdlr)

    # if skylog.ini is disabled or not available, log at least to stderr
    if not got_skylog:
        fstr = cf.get('logfmt_console', def_fmt)
        fstr_date = cf.get('logdatefmt_console', def_datefmt)
        if log_level < logging.INFO:
            fstr = cf.get('logfmt_console_verbose', fstr)
            fstr_date = cf.get('logdatefmt_console_verbose', fstr_date)
        stream_hdlr = logging.StreamHandler()
        fmt = logging.Formatter(fstr, fstr_date)
        stream_hdlr.setFormatter(fmt)
        root.addHandler(stream_hdlr)

    return log


class BaseScript:
    """Base class for service scripts.

    Handles logging, daemonizing, config, errors.

    Config template::

        ## Parameters for skytools.BaseScript ##

        # how many seconds to sleep between work loops
        # if missing or 0, then instead sleeping, the script will exit
        loop_delay = 1.0

        # where to log
        logfile = ~/log/%(job_name)s.log

        # where to write pidfile
        pidfile = ~/pid/%(job_name)s.pid

        # per-process name to use in logging
        #job_name = %(config_name)s

        # whether centralized logging should be used
        # search-path [ ./skylog.ini, ~/.skylog.ini, /etc/skylog.ini ]
        #   0 - disabled
        #   1 - enabled, unless non-daemon on console (os.isatty())
        #   2 - always enabled
        #use_skylog = 0

        # where to find skylog.ini
        #skylog_locations = skylog.ini, ~/.skylog.ini, /etc/skylog.ini

        # how many seconds to sleep after catching a exception
        #exception_sleep = 20
    """
    service_name: str
    job_name: str
    cf: "skytools.Config"
    go_daemon: bool
    need_reload: bool

    cf_defaults: Dict[str, str] = {}
    pidfile: Optional[str] = None

    # >0 - sleep time if work() requests sleep
    # 0  - exit if work requests sleep
    # <0 - run work() once [same as looping=0]
    loop_delay: float = 1.0

    # 0 - run work() once
    # 1 - run work() repeatedly
    looping: int = 1

    # result from last work() call:
    #  1 - there is probably more work, don't sleep
    #  0 - no work, sleep before calling again
    # -1 - exception was thrown
    work_state: int = 1

    # setup logger here, this allows override by subclass
    log = logging.getLogger('skytools.BaseScript')

    # start time
    started: float = 0

    # set to True to use argparse
    ARGPARSE: bool = False

    def __init__(self, service_name: str, args: Sequence[str]) -> None:
        """Script setup.

        User class should override work() and optionally __init__(), startup(),
        reload(), reset(), shutdown() and init_optparse().

        NB: In case of daemon, __init__() and startup()/work()/shutdown() will be
        run in different processes.  So nothing fancy should be done in __init__().

        @param service_name: unique name for script.
            It will be also default job_name, if not specified in config.
        @param args: cmdline args (sys.argv[1:]), but can be overridden
        """
        self.service_name = service_name
        self.go_daemon = False
        self.need_reload = False
        self.exception_count = 0
        self.stat_dict: Dict[str, float] = {}
        self.log_level = logging.INFO

        # parse command line
        self.options, self.args = self.parse_args(args)

        # check args
        if self.options.version:
            self.print_version()
            sys.exit(0)
        if self.options.daemon:
            self.go_daemon = True
        if self.options.quiet:
            self.log_level = logging.WARNING
        if self.options.verbose:
            if self.options.verbose > 1:
                self.log_level = skytools.skylog.TRACE
            else:
                self.log_level = logging.DEBUG

        self.cf_override = {}
        if self.options.set:
            for a in self.options.set:
                k, v = a.split('=', 1)
                self.cf_override[k.strip()] = v.strip()

        if self.options.ini:
            self.print_ini()
            sys.exit(0)

        # read config file
        self.reload()

        # init logging
        _init_log(self.job_name, self.service_name, self.cf, self.log_level, self.go_daemon)

        # send signal, if needed
        if self.options.cmd == "kill":
            self.send_signal(signal.SIGTERM)
        elif self.options.cmd == "stop":
            self.send_signal(signal.SIGINT)
        elif self.options.cmd == "reload":
            self.send_signal(signal.SIGHUP)

    def parse_args(self, args: Sequence[str]) -> Tuple[Any, Sequence[str]]:
        if self.ARGPARSE:
            arg_parser = self.init_argparse()
            options = arg_parser.parse_args(args)
            args = getattr(options, "args", [])
            return options, args
        opt_parser = self.init_optparse()
        options2, args2 = opt_parser.parse_args(args)
        return options2, args2

    def print_version(self) -> None:
        service = self.service_name
        ver = getattr(self, '__version__', None)
        if ver:
            service += ' version %s' % ver
        print('%s, Skytools version %s' % (service, getattr(skytools, '__version__')))

    def print_ini(self) -> None:
        """Prints out ini file from doc string of the script of default for dbscript

        Used by --ini option on command line.
        """

        # current service name
        print("[%s]\n" % self.service_name)

        # walk class hierarchy
        bases = [self.__class__]
        while len(bases) > 0:
            parents = []
            for c in bases:
                for p in c.__bases__:
                    if p not in parents:
                        parents.append(p)
                doc = c.__doc__
                if doc:
                    self._print_ini_frag(doc)
            bases = parents

    def _print_ini_frag(self, doc: str) -> None:
        # use last '::' block as config template
        pos = doc and doc.rfind('::\n') or -1
        if pos < 0:
            return
        doc = doc[pos + 2:].rstrip()
        doc = skytools.dedent(doc)

        # merge overrided options into output
        for ln in doc.splitlines():
            vals = ln.split('=', 1)
            if len(vals) != 2:
                print(ln)
                continue

            k = vals[0].strip()
            v = vals[1].strip()
            if k and k[0] == '#':
                print(ln)
                k = k[1:]
                if k in self.cf_override:
                    print('%s = %s' % (k, self.cf_override[k]))
            elif k in self.cf_override:
                if v:
                    print('#' + ln)
                print('%s = %s' % (k, self.cf_override[k]))
            else:
                print(ln)

        print('')

    def load_config(self) -> skytools.Config:
        """Loads and returns skytools.Config instance.

        By default it uses first command-line argument as config
        file name.  Can be overridden.
        """

        if len(self.args) < 1:
            print("need config file, use --help for help.")
            sys.exit(1)
        conf_file = self.args[0]
        return skytools.Config(self.service_name, conf_file,
                               user_defs=self.cf_defaults,
                               override=self.cf_override)

    def init_optparse(self, parser: Optional[optparse.OptionParser] = None) -> optparse.OptionParser:
        """Initialize a OptionParser() instance that will be used to
        parse command line arguments.

        Note that it can be overridden both directions - either DBScript
        will initialize an instance and pass it to user code or user can
        initialize and then pass to DBScript.init_optparse().

        @param parser: optional OptionParser() instance,
               where DBScript should attach its own arguments.
        @return: initialized OptionParser() instance.
        """
        if parser:
            p = parser
        else:
            p = optparse.OptionParser()
            p.set_usage("%prog [options] INI")

        # generic options
        p.add_option("-q", "--quiet", action="store_true",
                     help="log only errors and warnings")
        p.add_option("-v", "--verbose", action="count",
                     help="log verbosely")
        p.add_option("-d", "--daemon", action="store_true",
                     help="go background")
        p.add_option("-V", "--version", action="store_true",
                     help="print version info and exit")
        p.add_option("", "--ini", action="store_true",
                     help="display sample ini file")
        p.add_option("", "--set", action="append",
                     help="override config setting (--set 'PARAM=VAL')")

        # control options
        g = optparse.OptionGroup(p, 'control running process')
        g.add_option("-r", "--reload",
                     action="store_const", const="reload", dest="cmd",
                     help="reload config (send SIGHUP)")
        g.add_option("-s", "--stop",
                     action="store_const", const="stop", dest="cmd",
                     help="stop program safely (send SIGINT)")
        g.add_option("-k", "--kill",
                     action="store_const", const="kill", dest="cmd",
                     help="kill program immediately (send SIGTERM)")
        p.add_option_group(g)

        return p

    def init_argparse(self, parser: Optional[argparse.ArgumentParser] = None) -> argparse.ArgumentParser:
        """Initialize a ArgumentParser() instance that will be used to
        parse command line arguments.

        Note that it can be overridden both directions - either BaseScript
        will initialize an instance and pass it to user code or user can
        initialize and then pass to BaseScript.init_optparse().

        @param parser: optional ArgumentParser() instance,
               where BaseScript should attach its own arguments.
        @return: initialized ArgumentParser() instance.
        """
        if parser:
            p = parser
        else:
            p = argparse.ArgumentParser()

        # generic options
        p.add_argument("-q", "--quiet", action="store_true",
                       help="log only errors and warnings")
        p.add_argument("-v", "--verbose", action="count",
                       help="log verbosely")
        p.add_argument("-d", "--daemon", action="store_true",
                       help="go background")
        p.add_argument("-V", "--version", action="store_true",
                       help="print version info and exit")
        p.add_argument("--ini", action="store_true",
                       help="display sample ini file")
        p.add_argument("--set", action="append",
                       help="override config setting (--set 'PARAM=VAL')")
        p.add_argument("args", nargs="*")

        # control options
        g = p.add_argument_group('control running process')
        g.add_argument("-r", "--reload",
                       action="store_const", const="reload", dest="cmd",
                       help="reload config (send SIGHUP)")
        g.add_argument("-s", "--stop",
                       action="store_const", const="stop", dest="cmd",
                       help="stop program safely (send SIGINT)")
        g.add_argument("-k", "--kill",
                       action="store_const", const="kill", dest="cmd",
                       help="kill program immediately (send SIGTERM)")

        return p

    def send_signal(self, sig: int) -> None:
        if not self.pidfile:
            self.log.warning("No pidfile in config, nothing to do")
        elif os.path.isfile(self.pidfile):
            alive = skytools.signal_pidfile(self.pidfile, sig)
            if not alive:
                self.log.warning("pidfile exists, but process not running")
        else:
            self.log.warning("No pidfile, process not running")
        sys.exit(0)

    def set_single_loop(self, do_single_loop: int) -> None:
        """Changes whether the script will loop or not."""
        if do_single_loop:
            self.looping = 0
        else:
            self.looping = 1

    def _boot_daemon(self) -> None:
        run_single_process(self, self.go_daemon, self.pidfile)

    def start(self) -> None:
        """This will launch main processing thread."""
        if self.go_daemon:
            if not self.pidfile:
                self.log.error("Daemon needs pidfile")
                sys.exit(1)
        self.run_func_safely(self._boot_daemon)

    def stop(self) -> None:
        """Safely stops processing loop."""
        self.looping = 0

    def reload(self) -> None:
        "Reload config."
        # avoid double loading on startup
        if not getattr(self, "cf", None):
            self.cf = self.load_config()
        else:
            self.cf.reload()
            self.log.info("Config reloaded")
        self.job_name = self.cf.get("job_name")
        self.pidfile = self.cf.getfile("pidfile", '')
        self.loop_delay = self.cf.getfloat("loop_delay", self.loop_delay)
        self.exception_sleep = self.cf.getfloat("exception_sleep", 20)
        self.exception_quiet = self.cf.getlist("exception_quiet", [])
        self.exception_grace = self.cf.getfloat("exception_grace", 5 * 60)
        self.exception_reset = self.cf.getfloat("exception_reset", 15 * 60)

    def hook_sighup(self, sig: int, frame: Any) -> None:
        "Internal SIGHUP handler.  Minimal code here."
        self.need_reload = True

    last_sigint: float = 0

    def hook_sigint(self, sig: int, frame: Any) -> None:
        "Internal SIGINT handler.  Minimal code here."
        self.stop()
        t = time.time()
        if t - self.last_sigint < 1:
            self.log.warning("Double ^C, fast exit")
            sys.exit(1)
        self.last_sigint = t

    def stat_get(self, key: str) -> Optional[float]:
        """Reads a stat value."""
        try:
            return self.stat_dict[key]
        except KeyError:
            return None

    def stat_put(self, key: str, value: float) -> None:
        """Sets a stat value."""
        self.stat_dict[key] = value

    def stat_increase(self, key: str, increase: float = 1) -> None:
        """Increases a stat value."""
        try:
            self.stat_dict[key] += increase
        except KeyError:
            self.stat_dict[key] = increase

    def send_stats(self) -> None:
        "Send statistics to log."

        res = []
        for k, v in self.stat_dict.items():
            res.append("%s: %s" % (k, v))

        if len(res) == 0:
            return

        logmsg = "{%s}" % ", ".join(res)
        self.log.info(logmsg)
        self.stat_dict = {}

    def reset(self) -> None:
        "Something bad happened, reset all state."
        pass

    def run(self) -> None:
        "Thread main loop."

        # run startup, safely
        self.run_func_safely(self.startup)

        while True:
            # reload config, if needed
            if self.need_reload:
                self.reload()
                self.need_reload = False

            # do some work
            work = self.run_once()

            if not self.looping or self.loop_delay < 0:
                break

            # remember work state
            if work:
                self.work_state = 1 if work > 0 else -1
            else:
                self.work_state = 0
            # should sleep?
            if not work:
                if self.loop_delay > 0:
                    self.sleep(self.loop_delay)
                    if not self.looping:
                        break
                else:
                    break

        # run shutdown, safely?
        self.shutdown()

    def run_once(self) -> int:
        state = self.run_func_safely(self.work, True)

        # send stats that was added
        self.send_stats()

        return state

    last_func_fail: Optional[float] = None

    def run_func_safely(self, func: Callable[[], Optional[int]], prefer_looping: bool = False) -> int:
        "Run users work function, safely."
        try:
            r = func()
            if self.last_func_fail and time.time() > self.last_func_fail + self.exception_reset:
                self.last_func_fail = None
            # set exception count to 0 after success
            self.exception_count = 0
            return 1 if r else 0
        except UsageError as d:
            self.log.error(str(d))
            sys.exit(1)
        except MemoryError:
            try:  # complex logging may not succeed
                self.log.exception("Job %s out of memory, exiting", self.job_name)
            except MemoryError:
                self.log.fatal("Out of memory")
            sys.exit(1)
        except SystemExit as d:
            self.send_stats()
            if prefer_looping and self.looping and self.loop_delay > 0:
                self.log.info("got SystemExit(%s), exiting", str(d))
            self.reset()
            raise d
        except KeyboardInterrupt:
            self.send_stats()
            if prefer_looping and self.looping and self.loop_delay > 0:
                self.log.info("got KeyboardInterrupt, exiting")
            self.reset()
            sys.exit(1)
        except Exception as d:
            try:  # this may fail too
                self.send_stats()
            except BaseException:
                pass
            if self.last_func_fail is None:
                self.last_func_fail = time.time()
            emsg = str(d).rstrip()
            self.reset()
            self.exception_hook(d, emsg)
        # reset and sleep
        self.reset()
        if prefer_looping and self.looping and self.loop_delay > 0:
            # increase exception count & sleep
            self.exception_count += 1
            self.sleep_on_exception()
            return -1
        sys.exit(1)

    def sleep(self, secs: float) -> None:
        """Make script sleep for some amount of time."""
        try:
            time.sleep(secs)
        except IOError as ex:
            if ex.errno != errno.EINTR:
                raise

    def sleep_on_exception(self) -> None:
        """Make script sleep for some amount of time when an exception occurs.

        To implement more advance exception sleeping like exponential backoff you
        can override this method. Also note that you can use self.exception_count
        to track the number of consecutive exceptions.
        """
        self.sleep(self.exception_sleep)

    def _is_quiet_exception(self, ex: Exception) -> bool:
        if "ALL" in self.exception_quiet or ex.__class__.__name__ in self.exception_quiet:
            if self.last_func_fail and time.time() < self.last_func_fail + self.exception_grace:
                return True
        return False

    def exception_hook(self, det: Exception, emsg: str) -> None:
        """Called on after exception processing.

        Can do additional logging.

        @param det: exception details
        @param emsg: exception msg
        """
        lm = "Job %s crashed: %s" % (self.job_name, emsg)
        if self._is_quiet_exception(det):
            self.log.warning(lm)
        else:
            self.log.exception(lm)

    def work(self) -> Optional[int]:
        """Here should user's processing happen.

        Return value is taken as boolean - if true, the next loop
        starts immediately.  If false, DBScript sleeps for a loop_delay.
        """
        raise Exception("Nothing implemented?")

    def startup(self) -> None:
        """Will be called just before entering main loop.

        In case of daemon, if will be called in same process as work(),
        unlike __init__().
        """
        self.started = time.time()

        # set signals
        if hasattr(signal, 'SIGHUP'):
            signal.signal(signal.SIGHUP, self.hook_sighup)
        if hasattr(signal, 'SIGINT'):
            signal.signal(signal.SIGINT, self.hook_sigint)

    def shutdown(self) -> None:
        """Will be called just after exiting main loop.

        In case of daemon, if will be called in same process as work(),
        unlike __init__().
        """
        pass

    # define some aliases (short-cuts / backward compatibility cruft)
    stat_add = stat_put                 # Old, deprecated function.
    stat_inc = stat_increase


##
##  DBScript
##

#: how old connections need to be closed
DEF_CONN_AGE = 20 * 60  # 20 min


class DBScript(BaseScript):
    """Base class for database scripts.

    Handles database connection state.

    Config template::

        ## Parameters for skytools.DBScript ##

        # default lifetime for database connections (in seconds)
        #connection_lifetime = 1200
    """

    db_cache: Dict[str, "DBCachedConn"]
    _db_defaults: Dict[str, Mapping[str, int]]
    _listen_map: Dict[str, List[str]]

    def __init__(self, service_name: str, args: Sequence[str]) -> None:
        """Script setup.

        User class should override work() and optionally __init__(), startup(),
        reload(), reset() and init_optparse().

        NB: in case of daemon, the __init__() and startup()/work() will be
        run in different processes.  So nothing fancy should be done in __init__().

        @param service_name: unique name for script.
            It will be also default job_name, if not specified in config.
        @param args: cmdline args (sys.argv[1:]), but can be overridden
        """
        self.db_cache: Dict[str, DBCachedConn] = {}
        self._db_defaults = {}
        self._listen_map = {}  # dbname: channel_list
        super().__init__(service_name, args)

    def connection_hook(self, dbname: str, conn: Connection) -> None:
        pass

    def set_database_defaults(self, dbname: str, **kwargs: int) -> None:
        self._db_defaults[dbname] = kwargs

    def add_connect_string_profile(self, connstr: str, profile: Optional[str]) -> str:
        """Add extra profile info to connect string.
        """
        if profile:
            extra = self.cf.get("%s_extra_connstr" % profile, '')
            if extra:
                connstr += ' ' + extra
        return connstr

    def get_database(self, dbname: str, autocommit: int = 0, isolation_level: int = -1,
            cache: Optional[str] = None, connstr: Optional[str] = None,
            profile: Optional[str] = None) -> Connection:
        """Load cached database connection.

        User must not store it permanently somewhere,
        as all connections will be invalidated on reset.
        """

        max_age = self.cf.getint('connection_lifetime', DEF_CONN_AGE)

        if not cache:
            cache = dbname

        params: Dict[str, int] = {}
        defs = self._db_defaults.get(cache, {})
        for k in defs:
            params[k] = defs[k]
        if isolation_level >= 0:
            params['isolation_level'] = isolation_level
        elif autocommit:
            params['isolation_level'] = 0
        elif params.get('autocommit', 0):
            params['isolation_level'] = 0
        elif 'isolation_level' not in params:
            params['isolation_level'] = skytools.I_READ_COMMITTED

        if 'max_age' not in params:
            params['max_age'] = max_age

        if cache in self.db_cache:
            dbc = self.db_cache[cache]
            if connstr is None:
                connstr = self.cf.get(dbname, '')
            if connstr:
                connstr = self.add_connect_string_profile(connstr, profile)
                dbc.check_connstr(connstr)
        else:
            if not connstr:
                connstr = self.cf.get(dbname)
            connstr = self.add_connect_string_profile(connstr, profile)

            # connstr might contain password, it is not a good idea to log it
            filtered_connstr = connstr
            pos = connstr.lower().find('password')
            if pos >= 0:
                filtered_connstr = connstr[:pos] + ' [...]'

            self.log.debug("Connect '%s' to '%s'", cache, filtered_connstr)
            dbc = DBCachedConn(cache, connstr, params['max_age'], setup_func=self.connection_hook)
            self.db_cache[cache] = dbc

        clist = []
        if cache in self._listen_map:
            clist = self._listen_map[cache]

        return dbc.get_connection(params['isolation_level'], clist)

    def close_database(self, dbname: str) -> None:
        """Explicitly close a cached connection.

        Next call to get_database() will reconnect.
        """
        if dbname in self.db_cache:
            dbc = self.db_cache[dbname]
            dbc.reset()
            del self.db_cache[dbname]

    def reset(self) -> None:
        "Something bad happened, reset all connections."
        for dbc in self.db_cache.values():
            dbc.reset()
        self.db_cache = {}
        super().reset()

    def run_once(self) -> int:
        state = super().run_once()

        # reconnect if needed
        for dbc in self.db_cache.values():
            dbc.refresh()

        return state

    def exception_hook(self, det: Exception, emsg: str) -> None:
        """Log database and query details from exception."""
        curs = getattr(det, 'cursor', None)
        conn = getattr(curs, 'connection', None)
        if conn:
            # db connection
            sql = getattr(curs, 'query', None) or '?'
            if isinstance(sql, bytes):
                sql = sql.decode('utf8')
            if len(sql) > 200:  # avoid logging londiste huge batched queries
                sql = sql[:60] + " ..."
            lm = "Job %s got error on connection: %s.   Query: %s" % (
                self.job_name, emsg, sql)
            if self._is_quiet_exception(det):
                self.log.warning(lm)
            else:
                self.log.exception(lm)
        else:
            super().exception_hook(det, emsg)

    def sleep(self, secs: float) -> None:
        """Make script sleep for some amount of time."""
        fdlist = []
        for dbname in self._listen_map:
            if dbname not in self.db_cache:
                continue
            fd = self.db_cache[dbname].fileno()
            if fd is None:
                continue
            fdlist.append(fd)

        if not fdlist:
            return super().sleep(secs)

        try:
            if hasattr(select, 'poll'):
                p = select.poll()
                for fd in fdlist:
                    p.register(fd, select.POLLIN)
                p.poll(int(secs * 1000))
            else:
                select.select(fdlist, [], [], secs)
        except select.error:
            self.log.info('wait canceled')
        return None

    def _exec_cmd(self, curs: Cursor, sql: str, args: ExecuteParams, quiet: bool = False, prefix: Optional[str] = None) -> Tuple[bool, Sequence[DictRow]]:
        """Internal tool: Run SQL on cursor."""
        if self.options.verbose:
            self.log.debug("exec_cmd: %s", skytools.quote_statement(sql, args))

        _pfx = ""
        if prefix:
            _pfx = "[%s] " % prefix
        curs.execute(sql, args)
        ok = True
        rows = curs.fetchall()
        for row in rows:
            try:
                code = row['ret_code']
                msg = row['ret_note']
            except KeyError:
                self.log.error("Query does not conform to exec_cmd API:")
                self.log.error("SQL: %s", skytools.quote_statement(sql, args))
                self.log.error("Row: %s", repr(dict(row.items())))
                sys.exit(1)
            level = code // 100
            if level == 1:
                self.log.debug("%s%d %s", _pfx, code, msg)
            elif level == 2:
                if quiet:
                    self.log.debug("%s%d %s", _pfx, code, msg)
                else:
                    self.log.info("%s%s", _pfx, msg)
            elif level == 3:
                self.log.warning("%s%s", _pfx, msg)
            else:
                self.log.error("%s%s", _pfx, msg)
                self.log.debug("Query was: %s", skytools.quote_statement(sql, args))
                ok = False
        return (ok, rows)

    def _exec_cmd_many(self, curs: Cursor, sql: str, baseargs: List[Any], extra_list: Sequence[Any], quiet:bool=False, prefix:Optional[str]=None) -> Tuple[bool, Sequence[DictRow]]:
        """Internal tool: Run SQL on cursor multiple times."""
        ok = True
        rows: List[DictRow] = []
        for a in extra_list:
            (tmp_ok, tmp_rows) = self._exec_cmd(curs, sql, baseargs + [a], quiet, prefix)
            if not tmp_ok:
                ok = False
            rows += tmp_rows
        return (ok, rows)

    def exec_cmd(self, db_or_curs: Union[Connection, Cursor], q: str, args: ExecuteParams,
            commit: bool = True, quiet: bool = False, prefix: Optional[str] = None) -> Sequence[DictRow]:
        """Run SQL on db with code/value error handling."""

        db: Optional[Connection]
        curs: Cursor

        if hasattr(db_or_curs, 'cursor'):
            db = cast(Connection, db_or_curs)
            curs = db.cursor()
        else:
            db = None
            curs = db_or_curs
        (ok, rows) = self._exec_cmd(curs, q, args, quiet, prefix)
        if ok:
            if commit and db:
                db.commit()
            return rows
        else:
            if db:
                db.rollback()
            if self.options.verbose:
                raise Exception("db error")
            # error is already logged
            sys.exit(1)

    def exec_cmd_many(self, db_or_curs: Union[Connection, Cursor], sql: str, baseargs: List[Any], extra_list: Sequence[Any],
            commit: bool = True, quiet: bool = False, prefix: Optional[str] = None) -> Sequence[DictRow]:
        """Run SQL on db multiple times."""
        if hasattr(db_or_curs, 'cursor'):
            db = cast(Connection, db_or_curs)
            curs = db.cursor()
        else:
            db = None
            curs = db_or_curs
        (ok, rows) = self._exec_cmd_many(curs, sql, baseargs, extra_list, quiet, prefix)
        if ok:
            if commit and db:
                db.commit()
            return rows
        else:
            if db:
                db.rollback()
            if self.options.verbose:
                raise Exception("db error")
            # error is already logged
            sys.exit(1)

    def execute_with_retry(self, dbname: str, stmt: str, args: List[Any], exceptions: Optional[Sequence[Type[Exception]]] = None) -> Tuple[int, Cursor]:
        """ Execute SQL and retry if it fails.
        Return number of retries and current valid cursor, or raise an exception.
        """
        sql_retry = self.cf.getbool("sql_retry", False)
        sql_retry_max_count = self.cf.getint("sql_retry_max_count", 10)
        sql_retry_max_time = self.cf.getint("sql_retry_max_time", 300)
        sql_retry_formula_a = self.cf.getint("sql_retry_formula_a", 1)
        sql_retry_formula_b = self.cf.getint("sql_retry_formula_b", 5)
        sql_retry_formula_cap = self.cf.getint("sql_retry_formula_cap", 60)
        elist = tuple(exceptions) if exceptions else ()
        stime = time.time()
        tried = 0
        dbc: Optional[DBCachedConn] = None
        while True:
            try:
                if dbc is None:
                    if dbname not in self.db_cache:
                        self.get_database(dbname, autocommit=1)
                    dbc = self.db_cache[dbname]
                    if dbc.isolation_level != skytools.I_AUTOCOMMIT:
                        raise skytools.UsageError("execute_with_retry: autocommit required")
                else:
                    dbc.reset()
                curs = dbc.get_connection(dbc.isolation_level).cursor()
                curs.execute(stmt, args)
                break
            except elist as e:
                if not sql_retry or tried >= sql_retry_max_count or time.time() - stime >= sql_retry_max_time:
                    raise
                self.log.info("Job %s got error on connection %s: %s", self.job_name, dbname, e)
            except BaseException:
                raise
            # y = a + bx , apply cap
            y = sql_retry_formula_a + sql_retry_formula_b * tried
            if sql_retry_formula_cap is not None and y > sql_retry_formula_cap:
                y = sql_retry_formula_cap
            tried += 1
            self.log.info("Retry #%i in %i seconds ...", tried, y)
            self.sleep(y)
        return tried, curs

    def listen(self, dbname: str, channel: str) -> None:
        """Make connection listen for specific event channel.

        Listening will be activated on next .get_database() call.

        Basically this means that DBScript.sleep() will poll for events
        on that db connection, so when event appears, script will be
        woken up.
        """
        if dbname not in self._listen_map:
            self._listen_map[dbname] = []
        clist = self._listen_map[dbname]
        if channel not in clist:
            clist.append(channel)

    def unlisten(self, dbname: str, channel: str = '*') -> None:
        """Stop connection for listening on specific event channel.

        Listening will stop on next .get_database() call.
        """
        if dbname not in self._listen_map:
            return
        if channel == '*':
            del self._listen_map[dbname]
            return
        clist = self._listen_map[dbname]
        try:
            clist.remove(channel)
        except ValueError:
            pass


SetupFunc = Callable[[str, Connection], None]


class DBCachedConn:
    """Cache a db connection."""

    name: str
    loc: str
    conn: Optional[Connection]
    conn_time: float
    max_age: int
    isolation_level: int
    verbose: bool
    setup_func: Optional[SetupFunc]
    listen_channel_list: Sequence[str]

    def __init__(self, name: str, loc: str, max_age:int=DEF_CONN_AGE, verbose:bool=False, setup_func: Optional[SetupFunc] = None, channels: Sequence[str] = ()) -> None:
        self.name = name
        self.loc = loc
        self.conn = None
        self.conn_time = 0
        self.max_age = max_age
        self.isolation_level = -1
        self.verbose = verbose
        self.setup_func = setup_func
        self.listen_channel_list = []

    def fileno(self) -> Optional[int]:
        if not self.conn:
            return None
        return self.conn.cursor().fileno()

    def get_connection(self, isolation_level:int=-1, listen_channel_list: Sequence[str]=()) -> Connection:

        # default isolation_level is READ COMMITTED
        if isolation_level < 0:
            isolation_level = skytools.I_READ_COMMITTED

        # new conn?
        if not self.conn:
            self.isolation_level = isolation_level
            conn = skytools.connect_database(self.loc)
            conn = skytools.connect_database(self.loc)
            conn.set_isolation_level(isolation_level)

            self.conn = conn
            self.conn_time = time.time()
            if self.setup_func:
                self.setup_func(self.name, conn)
        else:
            if self.isolation_level != isolation_level:
                raise Exception("Conflict in isolation_level")
            conn = self.conn

        self._sync_listen(listen_channel_list)

        # done
        return conn

    def _sync_listen(self, new_clist: Sequence[str]) -> None:
        if not new_clist and not self.listen_channel_list:
            return
        if not self.conn:
            return
        curs = self.conn.cursor()
        for ch in self.listen_channel_list:
            if ch not in new_clist:
                curs.execute("UNLISTEN %s" % skytools.quote_ident(ch))
        for ch in new_clist:
            if ch not in self.listen_channel_list:
                curs.execute("LISTEN %s" % skytools.quote_ident(ch))
        if self.isolation_level != skytools.I_AUTOCOMMIT:
            self.conn.commit()
        self.listen_channel_list = new_clist[:]

    def refresh(self) -> None:
        if not self.conn:
            return
        #for row in self.conn.notifies():
        #    if row[0].lower() == "reload":
        #        self.reset()
        #        return
        if not self.max_age:
            return
        if time.time() - self.conn_time >= self.max_age:
            self.reset()

    def reset(self) -> None:
        if not self.conn:
            return

        # drop reference
        conn = self.conn
        self.conn = None
        self.listen_channel_list = []

        # close
        try:
            conn.close()
        except BaseException:
            pass

    def check_connstr(self, connstr: str) -> None:
        """Drop connection if connect string has changed.
        """
        if self.loc != connstr:
            self.reset()




"""File utilities
"""

# pylint:disable=unspecified-encoding

import errno
import os
import sys
from typing import Optional, Union

__all__ = ['write_atomic', 'signal_pidfile']


def write_atomic_unix(fn: str, data: Union[bytes, str], bakext: Optional[str] = None, mode: str = 'b') -> None:
    """Write file with rename."""

    if mode not in ['', 'b', 't']:
        raise ValueError("unsupported fopen mode")

    if mode == "b" and not isinstance(data, bytes):
        data = data.encode('utf8')

    mode = 'w' + mode

    # write new data to tmp file
    fn2 = fn + '.new'
    if "b" in mode:
        with open(fn2, mode) as f:
            f.write(data)
    else:
        with open(fn2, mode, encoding="utf8") as f:
            f.write(data)

    # link old data to bak file
    if bakext:
        if bakext.find('/') >= 0:
            raise ValueError("invalid bakext")
        fnb = fn + bakext
        try:
            os.unlink(fnb)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        try:
            os.link(fn, fnb)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    # win32 does not like replace
    if sys.platform == 'win32':
        try:
            os.remove(fn)
        except BaseException:
            pass

    # atomically replace file
    os.rename(fn2, fn)


def signal_pidfile(pidfile: str, sig: int) -> bool:
    """Send a signal to process whose ID is located in pidfile.

    Read only first line of pidfile to support multiline
    pidfiles like postmaster.pid.

    Returns True is successful, False if pidfile does not exist
    or process itself is dead.  Any other errors will passed
    as exceptions."""

    ln = ''
    try:
        with open(pidfile, 'r', encoding="utf8") as f:
            ln = f.readline().strip()
        pid = int(ln)
        if sig == 0 and sys.platform == 'win32':
            return win32_detect_pid(pid)
        os.kill(pid, sig)
        return True
    except (IOError, OSError) as ex:
        if ex.errno not in (errno.ESRCH, errno.ENOENT):
            raise
    except ValueError:
        # this leaves slight race when someone is just creating the file,
        # but more common case is old empty file.
        if not ln:
            return False
        raise ValueError('Corrupt pidfile: %s' % pidfile) from None
    return False


def win32_detect_pid(pid: int) -> bool:
    """Process detection for win32."""

    # avoid pywin32 dependecy, use ctypes instead
    import ctypes

    # win32 constants
    PROCESS_QUERY_INFORMATION = 1024
    STILL_ACTIVE = 259
    ERROR_INVALID_PARAMETER = 87
    ERROR_ACCESS_DENIED = 5

    # Load kernel32.dll
    k = ctypes.windll.kernel32  # type: ignore
    OpenProcess = k.OpenProcess
    OpenProcess.restype = ctypes.c_void_p

    # query pid exit code
    h = OpenProcess(PROCESS_QUERY_INFORMATION, 0, pid)
    if h is None:
        err = k.GetLastError()
        if err == ERROR_INVALID_PARAMETER:
            return False
        if err == ERROR_ACCESS_DENIED:
            return True
        raise OSError(errno.EFAULT, "Unknown win32error: " + str(err))
    code = ctypes.c_int()
    k.GetExitCodeProcess(h, ctypes.byref(code))
    k.CloseHandle(h)
    return code.value == STILL_ACTIVE


def win32_write_atomic(fn: str, data: Union[bytes, str], bakext: Optional[str] = None, mode: str = 'b') -> None:
    """Write file with rename for win32."""

    if mode not in ['', 'b', 't']:
        raise ValueError("unsupported fopen mode")

    if mode == "b" and not isinstance(data, bytes):
        data = data.encode('utf8')

    mode = "w" + mode

    # write new data to tmp file
    fn2 = fn + '.new'
    if "b" in mode:
        with open(fn2, mode) as f:
            f.write(data)
    else:
        with open(fn2, mode, encoding="utf8") as f:
            f.write(data)

    # move old data to bak file
    if bakext:
        if bakext.find('/') >= 0:
            raise ValueError("invalid bakext")
        fnb = fn + bakext
        try:
            os.remove(fnb)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        try:
            os.rename(fn, fnb)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
    else:
        try:
            os.remove(fn)
        except BaseException:
            pass

    # replace file
    os.rename(fn2, fn)


if sys.platform == 'win32':
    write_atomic = win32_write_atomic
else:
    write_atomic = write_atomic_unix




"""Setup for skytools module.
"""

from typing import Tuple
from setuptools import Extension, setup

try:
    from wheel.bdist_wheel import bdist_wheel
    class bdist_wheel_abi3(bdist_wheel):
        def get_tag(self) -> Tuple[str, str, str]:
            python, abi, plat = super().get_tag()
            if python.startswith("cp"):
                return CP_VER, "abi3", plat
            return python, abi, plat
    cmdclass = {"bdist_wheel": bdist_wheel_abi3}
except ImportError:
    cmdclass = {}

CP_VER = "cp37"
API_VER = ('Py_LIMITED_API', '0x03070000')

setup(
    cmdclass = cmdclass,
    ext_modules = [
        Extension("skytools._cquoting", ["modules/cquoting.c"],
                  define_macros=[API_VER], py_limited_api=True),
        Extension("skytools._chashtext", ["modules/hashtext.c"],
                  define_macros=[API_VER], py_limited_api=True),
    ]
)




"""
Implementation of Postgres hashing function.

hashtext_old() - used up to PostgreSQL 8.3
hashtext_new() - used since PostgreSQL 8.4
"""

import struct
import sys
from typing import Tuple, Union

try:
    from skytools._chashtext import hashtext_new, hashtext_old
except ImportError:
    def hashtext_old(v: Union[bytes, str]) -> int:
        return hashtext_old_py(v)
    def hashtext_new(v: Union[bytes, str]) -> int:
        return hashtext_new_py(v)


__all__ = ("hashtext_old", "hashtext_new")

# pad for last partial block
PADDING = b'\0' * 12


def uint32(x: int) -> int:
    """python does not have 32 bit integer so we need this hack to produce uint32 after bit operations"""
    return x & 0xffffffff


#
# Old Postgres hashtext() - lookup2 with custom initval
#

FMT_OLD = struct.Struct("<LLL")


def mix_old(a: int, b: int, c: int) -> Tuple[int, int, int]:
    c = uint32(c)

    a = uint32((a - b - c) ^ (c >> 13))
    b = uint32((b - c - a) ^ (a << 8))
    c = uint32((c - a - b) ^ (b >> 13))
    a = uint32((a - b - c) ^ (c >> 12))
    b = uint32((b - c - a) ^ (a << 16))
    c = uint32((c - a - b) ^ (b >> 5))
    a = uint32((a - b - c) ^ (c >> 3))
    b = uint32((b - c - a) ^ (a << 10))
    c = uint32((c - a - b) ^ (b >> 15))

    return a, b, c


def hashtext_old_py(k: Union[bytes, str]) -> int:
    """Old Postgres hashtext()"""

    if isinstance(k, str):
        k = k.encode()

    remain = len(k)
    pos = 0
    a = b = 0x9e3779b9
    c = 3923095

    # handle most of the key
    while remain >= 12:
        a2, b2, c2 = FMT_OLD.unpack_from(k, pos)
        a, b, c = mix_old(a + a2, b + b2, c + c2)
        pos += 12
        remain -= 12

    # handle the last 11 bytes
    a2, b2, c2 = FMT_OLD.unpack_from(k[pos:] + PADDING, 0)

    # the lowest byte of c is reserved for the length
    c2 = (c2 << 8) + len(k)

    a, b, c = mix_old(a + a2, b + b2, c + c2)

    # convert to signed int
    if c & 0x80000000:
        c = -0x100000000 + c

    return int(c)


#
# New Postgres hashtext() - hacked lookup3:
# - custom initval
# - calls mix() when len=12
# - shifted c in last block on little-endian
#

FMT_NEW = struct.Struct("=LLL")


def rol32(x: int, k: int) -> int:
    return ((x) << (k)) | (uint32(x) >> (32 - (k)))


def mix_new(a: int, b: int, c: int) -> Tuple[int, int, int]:
    a = (a - c) ^ rol32(c, 4)
    c += b
    b = (b - a) ^ rol32(a, 6)
    a += c
    c = (c - b) ^ rol32(b, 8)
    b += a
    a = (a - c) ^ rol32(c, 16)
    c += b
    b = (b - a) ^ rol32(a, 19)
    a += c
    c = (c - b) ^ rol32(b, 4)
    b += a

    return uint32(a), uint32(b), uint32(c)


def final_new(a: int, b: int, c: int) -> Tuple[int, int, int]:
    c = (c ^ b) - rol32(b, 14)
    a = (a ^ c) - rol32(c, 11)
    b = (b ^ a) - rol32(a, 25)
    c = (c ^ b) - rol32(b, 16)
    a = (a ^ c) - rol32(c, 4)
    b = (b ^ a) - rol32(a, 14)
    c = (c ^ b) - rol32(b, 24)

    return uint32(a), uint32(b), uint32(c)


def hashtext_new_py(k: Union[bytes, str]) -> int:
    """New Postgres hashtext()"""
    if isinstance(k, str):
        k = k.encode()
    remain = len(k)
    pos = 0
    a = b = c = 0x9e3779b9 + len(k) + 3923095

    # handle most of the key
    while remain >= 12:
        a2, b2, c2 = FMT_NEW.unpack_from(k, pos)
        a, b, c = mix_new(a + a2, b + b2, c + c2)
        pos += 12
        remain -= 12

    # handle the last 11 bytes
    a2, b2, c2 = FMT_NEW.unpack_from(k[pos:] + PADDING, 0)
    if sys.byteorder == 'little':
        c2 = c2 << 8
    a, b, c = final_new(a + a2, b + b2, c + c2)

    # convert to signed int
    if c & 0x80000000:
        c = -0x100000000 + c

    return int(c)





import gzip
import os
import tempfile

from skytools.gzlog import gzip_append


def test_gzlog() -> None:
    fd, tmpname = tempfile.mkstemp(suffix='.gz')
    os.close(fd)
    try:
        blk = b'1234567890' * 100
        write_total = 0
        for i in range(5):
            gzip_append(tmpname, blk)
            write_total += len(blk)

        read_total = 0
        with gzip.open(tmpname) as rfd:
            while True:
                blk = rfd.read(512)
                if not blk:
                    break
                read_total += len(blk)
    finally:
        os.remove(tmpname)

    assert read_total == write_total





from typing import Any

import pytest

from skytools.tnetstrings import dumps, loads


def ustr(v: Any) -> str:
    return repr(v).replace("u'", "'")


def nstr(b: bytes) -> str:
    return b.decode('utf8')


def test_dumps_simple_values() -> None:
    vals = (None, False, True, 222, 333.0, b"foo", u"bar")
    tnvals = [nstr(dumps(v)) for v in vals]
    assert tnvals == [
        '0:~', '5:false!', '4:true!', '3:222#', '5:333.0^', '3:foo,', '3:bar,'
    ]


def test_dumps_complex_values() -> None:
    vals2 = [[], (), {}, [1, 2], {'a': 'b'}]
    tnvals2 = [nstr(dumps(v)) for v in vals2]
    assert tnvals2 == ['0:]', '0:]', '0:}', '8:1:1#1:2#]', '8:1:a,1:b,}']


def test_loads_simple() -> None:
    vals = (None, False, True, 222, 333.0, b"foo", u"bar")
    res = ustr([loads(dumps(v)) for v in vals])
    assert res == "[None, False, True, 222, 333.0, 'foo', 'bar']"


def test_loads_complex() -> None:
    vals2 = [[], (), {}, [1, 2], {'a': 'b'}]
    res = ustr([loads(dumps(v)) for v in vals2])
    assert res == "[[], [], {}, [1, 2], {'a': 'b'}]"


def test_dumps_mview() -> None:
    res = nstr(dumps([memoryview(b'zzz'), b'qqq']))
    assert res == '12:3:zzz,3:qqq,]'


def test_loads_errors() -> None:
    with pytest.raises(ValueError):
        loads(b'4:qwez!')
    with pytest.raises(ValueError):
        loads(b'4:')
    with pytest.raises(ValueError):
        loads(b'4:qwez')
    with pytest.raises(ValueError):
        loads(b'4')
    with pytest.raises(ValueError):
        loads(b'')
    with pytest.raises(ValueError):
        loads(b'999999999999999999:z,')
    with pytest.raises(TypeError):
        loads(u'qweqwe')    # type: ignore[arg-type]
    with pytest.raises(ValueError):
        loads(b'4:true!0:~')
    with pytest.raises(ValueError):
        loads(b'1:X~')
    with pytest.raises(ValueError):
        loads(b'8:1:1#1:2#}')
    with pytest.raises(ValueError):
        loads(b'1:Xz')


def test_dumps_errors() -> None:
    with pytest.raises(TypeError):
        dumps(open)

